from moneywagon.core import Service
from moneywagon.crypto_data import crypto_data

class ShapeShiftIO(Service):

    api_homepage = "https://info.shapeshift.io/"

    def onchain_exchange_rates(self):
        pairs = self.get_url("https://shapeshift.io/marketinfo/").json()
        final_pairs = []

        for pair in pairs:
            deposit_code, withdraw_code = pair['pair'].split("_")
            try:
                deposit_name = crypto_data[deposit_code.lower()]['name']
            except (KeyError, TypeError): # currency not supported by moneywagon
                continue

            try:
                withdraw_name = crypto_data[withdraw_code.lower()]['name']
            except (KeyError, TypeError): # currency not supported by moneywagon
                continue

            final_pairs.append({
                'deposit_currency': {'code': deposit_code, 'name': deposit_name},
                'withdraw_currency': {'code': withdraw_code, 'name': withdraw_name},
                'rate': pair['rate'],
                'max_amount': pair['maxLimit'],
                'min_amount': pair['min'],
                'withdraw_fee': pair['minerFee'],
                'service': self
            })

        return final_pairs

    def onchain_exchange_status(self, deposit_address):
        return self.get_url("https://shapeshift.io/txStat/" + deposit_address).json()

    def get_onchain_exchange_address(self, deposit_crypto, withdraw_crypto, withdraw_address):
        url = "https://shapeshift.io/shift"
        pair = (deposit_crypto + "_" + withdraw_crypto).lower()
        return self.post_url(url, {'withdrawal': withdraw_address, 'pair': pair}).json()


# class MultiExplorerExchange(Service):
#
#     def onchain_exchange_rates(self):
#         return [
#         {
#         'deposit_currency': {'code': u'LTC', 'name': 'Litecoin'},
#         'max_amount': 508.36657379,
#         'min_amount': 0.09404389,
#         'rate': u'0.07631284',
#         'service': self,
#         'withdraw_currency': {'code': u'BTC', 'name': 'Bitcoin'},
#         'withdraw_fee': 0.0003}
#      ]


ALL_SERVICES = [ShapeShiftIO]



# -*- coding: utf-8 -*-
from __future__ import print_function

from binascii import unhexlify, hexlify
from hashlib import sha256
from unicodedata import normalize
import sys

from moneywagon.core import get_magic_bytes
from bitcoin import (
    privtopub, pubtoaddr, encode_privkey, get_privkey_format,
    encode_pubkey, changebase, fast_multiply, G, P, A, B
)

from base58 import b58encode_check, b58decode_check

try:
    import scrypt
except ImportError:
    raise ImportError("Scrypt is required for BIP38 support: pip install scrypt")

try:
    from Crypto.Cipher import AES
except ImportError:
    raise ImportError("PyCrypto is required for bip38 support: pip install pycrypto")

is_py2 = False
if sys.version_info <= (3,0):
    # py2
    is_py2 = True
else:
    # py3
    long = int
    unicode = str

def compress(x, y):
    """
    Given a x,y coordinate, encode in "compressed format"
    Returned is always 33 bytes.
    """
    polarity = "02" if y % 2 == 0 else "03"

    wrap = lambda x: x
    if not is_py2:
        wrap = lambda x: bytes(x, 'ascii')

    return unhexlify(wrap("%s%0.64x" % (polarity, x)))


def uncompress(payload):
    """
    Given a compressed ec key in bytes, uncompress it using math and return (x, y)
    """
    payload = hexlify(payload)
    even = payload[:2] == b"02"
    x = int(payload[2:], 16)
    beta = pow(int(x ** 3 + A * x + B), int((P + 1) // 4), int(P))
    y = (P-beta) if even else beta
    return x, y

def bytes_to_int(x):
    if is_py2:
        return int(x.encode('hex'), 16)
    else:
        return int.from_bytes(x, byteorder='big')


class Bip38Primitive(object):
    def __str__(self):
        return self.b58check


class Bip38EncryptedPrivateKey(Bip38Primitive):

    def __init__(self, crypto, b58check):
        if not b58check.startswith('6P'):
            raise Exception("Invalid encrypted private key. Must start wth 6P.")
        self.b58check = b58check
        self.payload = b58decode_check(b58check)
        second_byte = self.payload[1:2]
        self.flagbyte = self.payload[2:3]

        if second_byte == b'\x42':
            self.ec_multiply = False
            if self.flagbyte == b'\xC0':
                self.compressed = False
            elif self.flagbyte == b'\xE0':
                self.compressed = True
            else:
                raise Exception("Unknown flagbyte: %x" % flagbyte)

            self.addresshash = self.payload[3:7] # 4 bytes
            self.encryptedhalf1 = self.payload[7:23] # 16 bytes
            self.encryptedhalf2 = self.payload[23:39] # 16 bytes

        elif second_byte == b'\x43':
            self.ec_multiply = True
            if self.flagbyte == b'\x00':
                self.compressed = False
            elif self.flagbyte == b'\x20':
                self.compressed = True
            else:
                raise Exception("Unknown flagbyte: %x" % flagbyte)

            self.addresshash = self.payload[3:7] # 4 bytes
            self.ownerentropy = self.payload[7:15] # 8 bytes
            self.encryptedpart1 = self.payload[15:23] # 8 bytes
            self.encryptedpart2 = self.payload[23:] # 16 bytes

        else:
            raise Exception("Unknown second base58check byte: %x" % second_byte)

        self.pub_byte, self.priv_byte = get_magic_bytes(crypto)

    def decrypt(self, passphrase, wif=False):
        """
        BIP0038 non-ec-multiply decryption. Returns hex privkey.
        """
        passphrase = normalize('NFC', unicode(passphrase))
        if is_py2:
            passphrase = passphrase.encode('utf8')

        if self.ec_multiply:
            raise Exception("Not supported yet")

        key = scrypt.hash(passphrase, self.addresshash, 16384, 8, 8)
        derivedhalf1 = key[0:32]
        derivedhalf2 = key[32:64]

        aes = AES.new(derivedhalf2)
        decryptedhalf2 = aes.decrypt(self.encryptedhalf2)
        decryptedhalf1 = aes.decrypt(self.encryptedhalf1)
        priv = decryptedhalf1 + decryptedhalf2
        priv = unhexlify('%064x' % (long(hexlify(priv), 16) ^ long(hexlify(derivedhalf1), 16)))
        pub = privtopub(priv)

        if self.compressed:
            pub = encode_pubkey(pub, 'hex_compressed')
        addr = pubtoaddr(pub, self.pub_byte)

        if is_py2:
            ascii_key = addr
        else:
            ascii_key = bytes(addr,'ascii')

        if sha256(sha256(ascii_key).digest()).digest()[0:4] != self.addresshash:
            raise Exception('Bip38 password decrypt failed: Wrong password?')
        else:
            formatt = 'wif' if wif else 'hex'
            if self.compressed:
                return encode_privkey(priv, formatt + '_compressed', self.priv_byte)
            else:
                return encode_privkey(priv, formatt, self.priv_byte)

    @classmethod
    def encrypt(cls, crypto, privkey, passphrase):
        """
        BIP0038 non-ec-multiply encryption. Returns BIP0038 encrypted privkey.
        """
        pub_byte, priv_byte = get_magic_bytes(crypto)
        privformat = get_privkey_format(privkey)
        if privformat in ['wif_compressed','hex_compressed']:
            compressed = True
            flagbyte = b'\xe0'
            if privformat == 'wif_compressed':
                privkey = encode_privkey(privkey, 'hex_compressed')
                privformat = get_privkey_format(privkey)
        if privformat in ['wif', 'hex']:
            compressed = False
            flagbyte = b'\xc0'
        if privformat == 'wif':
            privkey = encode_privkey(privkey,'hex')
            privformat = get_privkey_format(privkey)

        pubkey = privtopub(privkey)
        addr = pubtoaddr(pubkey, pub_byte)

        passphrase = normalize('NFC', unicode(passphrase))
        if is_py2:
            ascii_key = addr
            passphrase = passphrase.encode('utf8')
        else:
            ascii_key = bytes(addr,'ascii')

        salt = sha256(sha256(ascii_key).digest()).digest()[0:4]
        key = scrypt.hash(passphrase, salt, 16384, 8, 8)
        derivedhalf1, derivedhalf2 = key[:32], key[32:]

        aes = AES.new(derivedhalf2)
        encryptedhalf1 = aes.encrypt(unhexlify('%0.32x' % (long(privkey[0:32], 16) ^ long(hexlify(derivedhalf1[0:16]), 16))))
        encryptedhalf2 = aes.encrypt(unhexlify('%0.32x' % (long(privkey[32:64], 16) ^ long(hexlify(derivedhalf1[16:32]), 16))))

        # 39 bytes    2 (6P)      1(R/Y)    4          16               16
        payload = b'\x01\x42' + flagbyte + salt + encryptedhalf1 + encryptedhalf2
        return cls(crypto, b58encode_check(payload))

    @classmethod
    def create_from_intermediate(cls, crypto, intermediate_point, seed, compressed=True, include_cfrm=True):
        """
        Given an intermediate point, given to us by "owner", generate an address
        and encrypted private key that can be decoded by the passphrase used to generate
        the intermediate point.
        """
        flagbyte = b'\x20' if compressed else b'\x00'
        payload = b58decode_check(str(intermediate_point))
        ownerentropy = payload[8:16]

        passpoint = payload[16:-4]
        x, y = uncompress(passpoint)

        if not is_py2:
            seed = bytes(seed, 'ascii')

        seedb = hexlify(sha256(seed).digest())[:24]
        factorb = int(hexlify(sha256(sha256(seedb).digest()).digest()), 16)
        generatedaddress = pubtoaddr(fast_multiply((x, y), factorb))

        wrap = lambda x: x
        if not is_py2:
            wrap = lambda x: bytes(x, 'ascii')

        addresshash = sha256(sha256(wrap(generatedaddress)).digest()).digest()[:4]

        encrypted_seedb = scrypt.hash(passpoint, addresshash + ownerentropy, 1024, 1, 1, 64)
        derivedhalf1, derivedhalf2 = encrypted_seedb[:32], encrypted_seedb[32:]

        aes = AES.new(derivedhalf2)
        block1 = long(seedb[0:16], 16) ^ long(hexlify(derivedhalf1[0:16]), 16)
        encryptedpart1 = aes.encrypt(unhexlify('%0.32x' % block1))

        block2 = long(hexlify(encryptedpart1[8:16]) + seedb[16:24], 16) ^ long(hexlify(derivedhalf1[16:32]), 16)
        encryptedpart2 = aes.encrypt(unhexlify('%0.32x' % block2))

        # 39 bytes      2           1           4              8                8                 16
        payload = b"\x01\x43" + flagbyte + addresshash + ownerentropy + encryptedpart1[:8] + encryptedpart2
        encrypted_pk = b58encode_check(payload)

        if not include_cfrm:
            return generatedaddress, encrypted_pk

        confirmation_code = Bip38ConfirmationCode.create(flagbyte, ownerentropy, factorb, derivedhalf1, derivedhalf2, addresshash)
        return generatedaddress, cls(crypto, encrypted_pk), confirmation_code


class Bip38IntermediatePoint(Bip38Primitive):

    lotseq_constant = 4096

    def __init__(self, b58check):
        self.lot = None
        self.sequence = None
        self.b58check = b58check
        if not b58check.startswith('passphrase'):
            raise Exception("Invalid intermediate point. Must start wth 'passphrase'.")
        payload = b58decode_check(str(b58check))
        self.ownerentropy = payload[8:16] # 8 bytes
        self.passpoint = payload[16:49] # 33 bytes

        flag = payload[7:8]

        if flag == b'\x51':
            self.has_lot_and_sequence = True
            self.ownersalt = self.ownerentropy[:4] # 4 bytes
            lotsequence = bytes_to_int(self.ownerentropy[4:]) # 4 bytes
            self.lot = lotsequence // self.lotseq_constant
            self.sequence = lotsequence % self.lot
        elif flag == b'\x53':
            self.has_lot_and_sequence = False
            self.ownersalt = self.ownerentropy # 8 bytes
        else:
            raise Exception("Unknown flag byte: %s, should be either \\x51 or \\x53" % flag)

    @classmethod
    def create(cls, passphrase, seed=None, ownersalt=None, lot=None, sequence=None):
        passphrase = normalize('NFC', unicode(passphrase))
        if is_py2:
            passphrase = passphrase.encode('utf8')

        if seed and not is_py2:
            seed = bytes(seed, 'ascii')

        if ownersalt and len(ownersalt) != 8:
            raise Exception("Invalid salt")

        if lot and sequence:
            if not ownersalt:
                ownersalt = sha256(seed).digest()[:4]
            lotseq = unhexlify("%0.8x" % (self.lotseq_constant * lot + sequence))
            ownerentropy = ownersalt + lotseq
        else:
            if not ownersalt:
                ownersalt = ownerentropy = sha256(seed).digest()[:8]
            else:
                ownerentropy = ownersalt

        prefactor = scrypt.hash(passphrase, ownersalt, 16384, 8, 8, 32)

        if lot and sequence:
            passfactor = sha256(sha256(prefactor + ownerentropy).digest()).digest()
        else:
            passfactor = prefactor

        passpoint = compress(*fast_multiply(G, bytes_to_int(passfactor)))

        last_byte = b'\x53' if not lot and not sequence else b'\x51'

        #                  7                            1            8            33
        payload = b'\x2C\xE9\xB3\xE1\xFF\x39\xE2' + last_byte + ownerentropy + passpoint
        return cls(b58encode_check(payload))


class Bip38ConfirmationCode(Bip38Primitive):

    def __init__(self, b58check):
        if not b58check.startswith('cfrm38'):
            raise Exception("Invalid Confirm code, must start with 'cfrm38'")
        self.b58check = b58check
        payload = b58decode_check(b58check)
        flagbyte = payload[5:6]

        if flagbyte == b'\x20':
            self.compressed = True
            self.sequence_and_lot = False
        elif flagbyte == b'\x00':
            self.compressed = False
            self.sequence_and_lot = False
        elif flagbyte == b'\x04':
            self.compressed = False
            self.sequence_and_lot = True
        elif flagbyte == b'\x24':
            self.compressed = True
            self.sequence_and_lot = True
        else:
            raise Exception("Unknown flagbyte: %s" % flagbyte)

        self.addresshash = payload[6:10] # 4 bytes
        self.ownerentropy = payload[10:18] # 8 bytes
        encryptedpointb = payload[18:] # 33 bytes

        self.pointbprefix = encryptedpointb[0:1] # 1 byte
        self.pointbx1 = encryptedpointb[1:17] # 16 bytes
        self.pointbx2 = encryptedpointb[17:] # 16 bytes

        if self.sequence_and_lot:
            self.ownersalt = self.ownerentropy[:4]
            lotsequence = bytes_to_int(self.ownerentropy[4:]) # 4 bytes
            self.lot = lotsequence // Bip38IntermediatePoint.lotseq_constant
            self.sequence = lotsequence % self.lot
        else:
            self.sequence, self.lot = None, None
            self.ownersalt = self.ownerentropy

    @classmethod
    def create(cls, flagbyte, ownerentropy, factorb, derivedhalf1, derivedhalf2, addresshash):
        pointb = compress(*fast_multiply(G, factorb))
        pointbprefix = bytearray([ord(pointb[:1]) ^ (ord(derivedhalf2[-1:]) & 1)])

        aes = AES.new(derivedhalf2)
        block1 = long(hexlify(pointb[1:17]), 16) ^ long(hexlify(derivedhalf1[:16]), 16)

        pointbx1 = aes.encrypt(unhexlify("%0.32x" % block1))
        block2 = long(hexlify(pointb[17:]), 16) ^ long(hexlify(derivedhalf1[16:]), 16)
        pointbx2 = aes.encrypt(unhexlify("%0.32x" % block2))

        #  33 bytes           1            16         16
        encryptedpointb = pointbprefix + pointbx1 + pointbx2

        #           5 (cfrm38 prefix)          1            4             8               33
        payload = b'\x64\x3B\xF6\xA8\x9A' + flagbyte + addresshash + ownerentropy + encryptedpointb
        return cls(b58encode_check(payload))

    def generate_address(self, passphrase):
        """
        Make sure the confirm code is valid for the given password and address.
        """

        inter = Bip38IntermediatePoint.create(passphrase, ownersalt=self.ownersalt)

        public_key = privtopub(inter.passpoint)

        # from Bip38EncryptedPrivateKey.create_from_intermediate
        derived = scrypt.hash(inter.passpoint, self.addresshash + inter.ownerentropy, 1024, 1, 1, 64)
        derivedhalf1, derivedhalf2 = derived[:32], derived[32:]

        unencrypted_prefix = bytes_to_int(self.pointbprefix) ^ (bytes_to_int(derived[63]) & 0x01);

        aes = AES.new(derivedhalf2)
        block1 = aes.decrypt(self.pointbx1)
        block2 = aes.decrypt(self.pointbx2)

        raise Exception("Not done yet")

        return
        block2 = long(hexlify(pointb2), 16) ^ long(hexlify(derivedhalf1[16:]), 16)


        return pubtoaddr(*fast_multiply(pointb, passfactor))



## tests below


def ec_test():
    """
    Tests the ec-multiply sections. Cases taken from the bip38 document.
    """
    seed = "dh3409sjgh3g48"
    seed2 = "asdas8729akjbmn"

    cases = [[
        'btc',
        'TestingOneTwoThree',
        'passphrasepxFy57B9v8HtUsszJYKReoNDV6VHjUSGt8EVJmux9n1J3Ltf1gRxyDGXqnf9qm',
        '6PfQu77ygVyJLZjfvMLyhLMQbYnu5uguoJJ4kMCLqWwPEdfpwANVS76gTX',
        '1PE6TQi6HTVNz5DLwB1LcpMBALubfuN2z2',
        '5K4caxezwjGCGfnoPTZ8tMcJBLB7Jvyjv4xxeacadhq8nLisLR2',
        ],[
        'btc',
        'Satoshi',
        'passphraseoRDGAXTWzbp72eVbtUDdn1rwpgPUGjNZEc6CGBo8i5EC1FPW8wcnLdq4ThKzAS',
        '6PfLGnQs6VZnrNpmVKfjotbnQuaJK4KZoPFrAjx1JMJUa1Ft8gnf5WxfKd',
        '1CqzrtZC6mXSAhoxtFwVjz8LtwLJjDYU3V',
        '5KJ51SgxWaAYR13zd9ReMhJpwrcX47xTJh2D3fGPG9CM8vkv5sH',
    ]]

    i = 1
    for crypto, passphrase, inter_point, encrypted_pk, address, decrypted_pk in cases:
        test_inter_point = Bip38IntermediatePoint.create(passphrase, seed)
        test_generated_address, test_encrypted_pk, test_cfm = Bip38EncryptedPrivateKey.create_from_intermediate(crypto, test_inter_point, seed2)
        assert test_generated_address == test_cfm.generate_address(passphrase), "Generate address from confirm code failed."
        print("EC multiply test #%s passed!" % i)
        i+= 1


    cases2 = [[
        'btc',
        'MOLON LABE',
        'passphraseaB8feaLQDENqCgr4gKZpmf4VoaT6qdjJNJiv7fsKvjqavcJxvuR1hy25aTu5sX',
        '6PgNBNNzDkKdhkT6uJntUXwwzQV8Rr2tZcbkDcuC9DZRsS6AtHts4Ypo1j',
        '1Jscj8ALrYu2y9TD8NrpvDBugPedmbj4Yh',
        '5JLdxTtcTHcfYcmJsNVy1v2PMDx432JPoYcBTVVRHpPaxUrdtf8',
        'cfrm38V8aXBn7JWA1ESmFMUn6erxeBGZGAxJPY4e36S9QWkzZKtaVqLNMgnifETYw7BPwWC9aPD',
        263183,
        1
        ],[
        'btc',
        u'ΜΟΛΩΝ ΛΑΒΕ',
        'passphrased3z9rQJHSyBkNBwTRPkUGNVEVrUAcfAXDyRU1V28ie6hNFbqDwbFBvsTK7yWVK',
        '6PgGWtx25kUg8QWvwuJAgorN6k9FbE25rv5dMRwu5SKMnfpfVe5mar2ngH',
        '1Lurmih3KruL4xDB5FmHof38yawNtP9oGf',
        '5KMKKuUmAkiNbA3DazMQiLfDq47qs8MAEThm4yL8R2PhV1ov33D',
        'cfrm38V8G4qq2ywYEFfWLD5Cc6msj9UwsG2Mj4Z6QdGJAFQpdatZLavkgRd1i4iBMdRngDqDs51',
        806938,
        1
    ]]

    i = 3
    for crypto, passphrase, inter_point, encrypted_pk, address, decrypted_pk, confirm, lot, sequence in cases2:
        test_inter_point = Bip38IntermediatePoint.create(passphrase, seed)
        test_generated_address, test_encrypted_pk, test_cfm = Bip38EncryptedPrivateKey.create_from_intermediate(crypto, inter_point, seed2)
        cfrm = Bip38ConfirmationCode(confirm)
        print("EC multiply test #%s passed!" % i)
        i += 1


def non_ec_test():
    # taken directly from the BIP38 whitepaper
    cases = [[
        'btc',
        '6PRVWUbkzzsbcVac2qwfssoUJAN1Xhrg6bNk8J7Nzm5H7kxEbn2Nh2ZoGg',
        'cbf4b9f70470856bb4f40f80b87edb90865997ffee6df315ab166d713af433a5',
        u'TestingOneTwoThree',
        False
        ], [
        'btc',
        '6PRNFFkZc2NZ6dJqFfhRoFNMR9Lnyj7dYGrzdgXXVMXcxoKTePPX1dWByq',
        '09c2686880095b1a4c249ee3ac4eea8a014f11e6f986d0b5025ac1f39afbd9ae',
        u'Satoshi',
        False
        ],[
        'btc',
        '6PRW5o9FLp4gJDDVqJQKJFTpMvdsSGJxMYHtHaQBF3ooa8mwD69bapcDQn',
        '5Jajm8eQ22H3pGWLEVCXyvND8dQZhiQhoLJNKjYXk9roUFTMSZ4',
        u'\u03D2\u0301\u0000\U00010400\U0001F4A9',
        True,
        ],[
        'btc',
        '6PYNKZ1EAgYgmQfmNVamxyXVWHzK5s6DGhwP4J5o44cvXdoY7sRzhtpUeo',
        'L44B5gGEpqEDRS9vVPz7QT35jcBG2r3CZwSwQ4fCewXAhAhqGVpP',
        u'TestingOneTwoThree',
        True
        ],[
        'btc',
        '6PYLtMnXvfG3oJde97zRyLYFZCYizPU5T3LwgdYJz1fRhh16bU7u6PPmY7',
        'KwYgW8gcxj1JWJXhPSu4Fqwzfhp5Yfi42mdYmMa4XqK7NJxXUSK7',
        u'Satoshi',
        True
    ]]

    index = 1
    for crypto, encrypted_key, unencrypted_key, password, use_wif in cases:
        pk = Bip38EncryptedPrivateKey.encrypt(crypto, unencrypted_key, password)
        assert encrypted_key == str(pk), "encrypt failed"
        assert unencrypted_key == pk.decrypt(password, wif=use_wif), 'decrypt failed'
        print("Non-ec multiply test #%s passed" % index)
        index += 1


def test():
    ec_test()
    #non_ec_test()



from __future__ import print_function
import sys
from binascii import hexlify
from tabulate import tabulate

from base58 import b58decode_check

from .core import (
    AutoFallbackFetcher, enforce_service_mode, get_optimal_services, get_magic_bytes, RevertToPrivateMode
)
from .historical_price import Quandl
from .crypto_data import crypto_data
from bitcoin import sha256, pubtoaddr, privtopub, encode_privkey, encode_pubkey

is_py2 = False
if sys.version_info <= (3,0):
    is_py2 = True


def get_current_price(crypto, fiat, services=None, **modes):
    if not services:
        services = get_optimal_services(crypto, 'current_price')

    return enforce_service_mode(
        services, CurrentPrice, {'crypto': crypto, 'fiat': fiat}, modes=modes
    )


def get_address_balance(crypto, address=None, addresses=None, services=None, **modes):
    if not services:
        services = get_optimal_services(crypto, 'address_balance')

    args = {'crypto': crypto}

    if address:
        args['address'] = address
    elif addresses:
        args['addresses'] = addresses
    else:
        raise Exception("Either address or addresses but not both")

    results = enforce_service_mode(
        services, AddressBalance, args, modes=modes
    )

    if addresses and 'total_balance' not in results:
        results['total'] = sum(results.values())

    if modes.get('private') and modes.get('report_services', False):
        # private mode does not return services (its not practical),
        # an empty list is returned in its place to simplify the API.
        return [], results

    return results

def get_historical_transactions(crypto, address=None, addresses=None, services=None, **modes):
    if not services:
        services = get_optimal_services(crypto, 'historical_transactions')

    kwargs = {'crypto': crypto}
    if addresses:
        kwargs['addresses'] = addresses
    if address:
        kwargs['address'] = address

    try:
        txs = enforce_service_mode(
            services, HistoricalTransactions, kwargs, modes=modes
        )
    except RevertToPrivateMode:
        # no services implement get_historical_transactions_multi...
        modes['private'] = 1
        if modes.get('verbose'):
            print("Can't make with single API call. Retrying with private mode")
        txs = enforce_service_mode(
            services, HistoricalTransactions, kwargs, modes=modes
        )

    if modes.get('private'):
        # private mode returns items indexed by address, this only makes sense to do
        # for address balance, so remove it here
        just_txs = []
        [just_txs.extend(x) for x in txs.values()]
        txs = sorted(just_txs, key=lambda tx: tx['date'], reverse=True)

        no_duplicates = []
        all_txids = []
        for tx in txs:
            # private mode may return duplicate txs, remove them here.
            if tx['txid'] in all_txids:
                continue
            all_txids.append(tx['txid'])
            no_duplicates.append(tx)

        txs = no_duplicates

        if modes.get('report_services', False):
            # private mode does not return services (its not practical),
            # an empty list is returned in its place to simplify the API.
            return [], txs



    return txs

def get_single_transaction(crypto, txid, services=None, **modes):
    if not services:
        services = get_optimal_services(crypto, 'single_transaction')

    return enforce_service_mode(
        services, SingleTransaction, {'crypto': crypto, 'txid': txid}, modes=modes
    )


def get_unspent_outputs(crypto, address=None, addresses=None, services=None, **modes):
    if not services:
        services = get_optimal_services(crypto, 'unspent_outputs')

    kwargs = {'crypto': crypto}
    if addresses:
        kwargs['addresses'] = addresses
    if address:
        kwargs['address'] = address

    try:
        utxos = enforce_service_mode(
            services, UnspentOutputs, kwargs, modes=modes
        )
    except RevertToPrivateMode:
        # no services implement get_unspent_outputs_multi...
        modes['private'] = 1
        if modes.get('verbose'):
            print("Can't make with single API call. Retrying with private mode")
        utxos = enforce_service_mode(
            services, UnspentOutputs, kwargs, modes=modes
        )

    if modes.get('private'):
        # private mode returns items indexed by address, this only makes sense to do
        # for address balance, so remove it here
        just_utxos = []
        [just_utxos.extend(x) for x in utxos.values()]
        utxos = sorted(just_utxos, key=lambda tx: tx['output'])
        if modes.get('report_services', False):
            # private mode does not return services (its not practical),
            # an empty list is returned in its place to satisfy the API.
            return [], utxos

    return utxos


def get_historical_price(crypto, fiat, date):
    """
    Only one service is defined for geting historical price, so no fetching modes
    are needed.
    """
    return HistoricalPrice().action(crypto, fiat, date)


def push_tx(crypto, tx_hex, services=None, **modes):
    if not services:
        services = get_optimal_services(crypto, 'push_tx')
    return enforce_service_mode(
        services, PushTx, {'crypto': crypto, 'tx_hex': tx_hex}, modes=modes
    )


def get_block(crypto, block_number='', block_hash='', latest=False, services=None, **modes):
    if not services:
        services = get_optimal_services(crypto, 'get_block')
    kwargs = dict(crypto=crypto, block_number=block_number, block_hash=block_hash, latest=latest)
    return enforce_service_mode(
        services, GetBlock, kwargs, modes=modes
    )


def get_optimal_fee(crypto, tx_bytes, **modes):
    """
    Get the optimal fee based on how big the transaction is. Currently this
    is only provided for BTC. Other currencies will return $0.02 in satoshi.
    """
    if crypto == 'btc':
        services = get_optimal_services(crypto, 'get_optimal_fee')
        fee = enforce_service_mode(
            services, OptimalFee, dict(crypto=crypto, tx_bytes=tx_bytes), modes=modes
        )
        if modes.get('report_services'):
            return fee[0], int(fee[1])
        else:
            return int(fee)
    else:
        convert = get_current_price(crypto, 'usd')
        fee = int(0.02 / convert * 1e8)

        if modes.get('report_services'):
            return [None], fee
        else:
            return fee


def get_onchain_exchange_rates(deposit_crypto=None, withdraw_crypto=None, **modes):
    """
    Gets exchange rates for all defined on-chain exchange services.
    """
    from moneywagon.onchain_exchange import ALL_SERVICES

    rates = []
    for Service in ALL_SERVICES:
        srv = Service(verbose=modes.get('verbose', False))
        rates.extend(srv.onchain_exchange_rates())

    if deposit_crypto:
        rates = [x for x in rates if x['deposit_currency']['code'] == deposit_crypto.upper()]

    if withdraw_crypto:
        rates = [x for x in rates if x['withdraw_currency']['code'] == withdraw_crypto.upper()]

    if modes.get('best', False):
        return max(rates, key=lambda x: float(x['rate']))

    return rates


def generate_keypair(crypto, seed, password=None):
    """
    Generate a private key and publickey for any currency, given a seed.
    That seed can be random, or a brainwallet phrase.
    """
    pub_byte, priv_byte = get_magic_bytes(crypto)
    priv = sha256(seed)
    pub = privtopub(priv)

    priv_wif = encode_privkey(priv, 'wif_compressed', vbyte=priv_byte)
    if password:
        # pycrypto etc. must be installed or this will raise ImportError, hence inline import.
        from .bip38 import Bip38EncryptedPrivateKey
        priv_wif = str(Bip38EncryptedPrivateKey.encrypt(crypto, priv_wif, password))

    compressed_pub = encode_pubkey(pub, 'hex_compressed')
    ret = {
        'public': {
            'hex_uncompressed': pub,
            'hex': compressed_pub,
            'address': pubtoaddr(compressed_pub, pub_byte)
        },
        'private': {
            'wif': priv_wif
        }
    }
    if not password:
        # only these are valid when no bip38 password is supplied
        ret['private']['hex'] = encode_privkey(priv, 'hex_compressed', vbyte=priv_byte)
        ret['private']['hex_uncompressed'] = encode_privkey(priv, 'hex', vbyte=priv_byte)
        ret['private']['wif_uncompressed'] = encode_privkey(priv, 'wif', vbyte=priv_byte)

    return ret


def sweep(crypto, private_key, to_address, fee=None, password=None, **modes):
    """
    Move all funds by private key to another address.
    """
    from moneywagon.tx import Transaction
    tx = Transaction(crypto, verbose=modes.get('verbose', False))
    tx.add_inputs(private_key=private_key, password=password, **modes)
    tx.change_address = to_address
    tx.fee(fee)

    return tx.push()


def get_explorer_url(crypto, address=None, txid=None, blocknum=None, blockhash=None):
    services = crypto_data[crypto]['services']['address_balance']
    urls = []
    context = {'crypto': crypto}
    if address:
        attr = "explorer_address_url"
        context['address'] = address
    elif txid:
        attr = "explorer_tx_url"
        context['txid'] = txid
    elif blocknum:
        attr = "explorer_blocknum_url"
        context['blocknum'] = blocknum
    elif blockhash:
        attr = "explorer_blockhash_url"
        context['blockhash'] = blockhash

    for service in services:
        template = getattr(service, attr)
        context['domain'] = service.domain
        context['protocol'] = service.protocol

        if hasattr(service, '_get_coin'):
            # used for when a service uses another name for a certain coin
            # other than the standard three letter currency code.
            context['coin'] = service._get_coin(crypto)

        if template:
            # render the explorer url temlate
            urls.append(template.format(**context))

    return urls


def guess_currency_from_address(address):
    """
    Given a crypto address, find which currency it likely belongs to.
    Raises an exception if it can't find a match. Raises exception if address
    is invalid.
    """
    if is_py2:
        fixer = lambda x: int(x.encode('hex'), 16)
    else:
        fixer = lambda x: x # does nothing

    first_byte = fixer(b58decode_check(address)[0])
    hits = []
    for currency, data in crypto_data.items():
        if hasattr(data, 'get'): # skip incomplete data listings
            version = data.get('address_version_byte', None)
            if version is not None and first_byte == version:
                hits.append([currency, data['name']])

    if hits:
        return hits

    raise ValueError("Unknown Currency with first byte: %s" % first_byte)


class OptimalFee(AutoFallbackFetcher):
    def action(self, crypto, tx_bytes):
        crypto = crypto.lower()
        return self._try_services("get_optimal_fee", crypto, tx_bytes)

    def no_service_msg(self, crypto, tx_bytes):
        return "Could not get optimal fee for: %s" % crypto


class SingleTransaction(AutoFallbackFetcher):
    def action(self, crypto, txid):
        crypto = crypto.lower()
        return self._try_services("get_single_transaction", crypto, txid)

    @classmethod
    def strip_for_consensus(cls, result):
        return "%.8f %.8f" % (result['total_in'], result['total_out'])

    def no_service_msg(self, crypto, txid=None, txids=None):
        return "Could not get transaction info for: %s:%s" % (crypto, txid or ', '.join(txids))


class GetBlock(AutoFallbackFetcher):
    def action(self, crypto, block_number='', block_hash='', latest=False):
        if sum([type(block_number)==int, bool(block_hash), bool(latest)]) != 1:
            raise ValueError("Only one of `block_hash`, `latest`, or `block_number` allowed.")
        return self._try_services(
            'get_block', crypto, block_number=block_number, block_hash=block_hash, latest=latest
        )

    def no_service_msg(self, crypto, block_number='', block_hash='', latest=False):
        return "Could not get %s block: %s%s%s" % (
            crypto, block_number, block_hash, 'latest' if latest else ''
        )

    @classmethod
    def strip_for_consensus(self, result):
        return "%s, %s, %s" % (
            result['hash'], result['block_number'], result['size']
        )


class HistoricalTransactions(AutoFallbackFetcher):
    def action(self, crypto, address=None, addresses=None):
        if addresses:
            method_name = "get_transactions_multi"
            kwargs = dict(addresses=addresses)

        if address:
            method_name = "get_transactions"
            kwargs = dict(address=address)

        txs = self._try_services(method_name, crypto, **kwargs)
        return sorted(txs, key=lambda tx: tx['date'], reverse=True)

    def no_service_msg(self, crypto, address=None, addresses=None):
        return "Could not get transactions for: %s:%s" % (crypto, address or ', '.join(addresses))

    @classmethod
    def strip_for_consensus(cls, results):
        stripped = []
        for result in results:
            result.sort(key=lambda x: x['date'])
            stripped.append(
                ", ".join(
                    ["[id: %s, amount: %s]" % (x['txid'], x['amount']) for x in result]
                )
            )
        return stripped


class UnspentOutputs(AutoFallbackFetcher):
    def action(self, crypto, address=None, addresses=None):
        if addresses:
            method_name = "get_unspent_outputs_multi"
            kwargs = dict(addresses=addresses)

        if address:
            method_name = "get_unspent_outputs"
            kwargs = dict(address=address)

        utxos = self._try_services(method_name, crypto=crypto, **kwargs)
        return sorted(utxos, key=lambda x: x['output'])

    def no_service_msg(self, crypto, address=None, addresses=None):
        return "Could not get unspent outputs for: %s:%s" % (crypto, address or ', '.join(addresses))

    @classmethod
    def strip_for_consensus(cls, results):
        stripped = []
        for result in results:
            result.sort(key=lambda x: x['output'])
            stripped.append(
                ", ".join(
                    ["[output: %s, value: %s]" % (x['output'], x['amount']) for x in result]
                )
            )
        return stripped


class CurrentPrice(AutoFallbackFetcher):
    def action(self, crypto, fiat):
        if crypto.lower() == fiat.lower():
            return (1.0, 'math')

        return self._try_services('get_current_price', crypto=crypto, fiat=fiat)

    def simplify_for_average(self, value):
        return value

    def no_service_msg(self, crypto, fiat):
        return "Can not find current price for %s->%s" % (crypto, fiat)


class AddressBalance(AutoFallbackFetcher):
    def action(self, crypto, address=None, addresses=None, confirmations=1):
        kwargs = dict(crypto=crypto, confirmations=confirmations)

        if address:
            method_name = "get_balance"
            kwargs['address'] = address

        if addresses:
            method_name = "get_balance_multi"
            kwargs['addresses'] = addresses

        results = self._try_services(method_name, **kwargs)

        if addresses and 'total_balance' not in results:
            results['total_balance'] = sum(results.values())

        return results

    def no_service_msg(self, crypto, address=None, addresses=None, confirmations=1):
        return "Could not get confirmed address balance for: %s" % crypto


class PushTx(AutoFallbackFetcher):
    def action(self, crypto, tx_hex):
        return self._try_services("push_tx", crypto=crypto, tx_hex=tx_hex)

    def no_service_msg(self, crypto, tx_hex):
        return "Could not push this %s transaction." % crypto


class HistoricalPrice(object):
    """
    This one doesn't inherit from AutoFallbackFetcher because there is only one
    historical price API service at the moment.
    """
    def __init__(self, responses=None, verbose=False):
        self.service = Quandl(responses, verbose=verbose)

    def action(self, crypto, fiat, at_time):
        crypto = crypto.lower()
        fiat = fiat.lower()

        if crypto != 'btc' and fiat != 'btc':
            # two external requests and some math is going to be needed.
            from_btc, source1, date1 = self.service.get_historical(crypto, 'btc', at_time)
            to_altcoin, source2, date2 = self.service.get_historical('btc', fiat, at_time)
            return (from_btc * to_altcoin), "%s x %s" % (source1, source2), date1
        else:
            return self.service.get_historical(crypto, fiat, at_time)

    @property
    def responses(self):
        return self.service.responses


def _get_all_services(crypto=None):
    """
    Go through the crypto_data structure and return all list of all (unique)
    installed services. Optionally filter by crypto-currency.
    """
    if not crypto:
        # no currency specified, get all services
        to_iterate = crypto_data.items()
    else:
        # limit to one currency
        to_iterate = [(crypto, crypto_data[crypto])]

    services = []
    for currency, data in to_iterate:
        if hasattr(data, 'append'):
            continue
        if 'services' not in data:
            continue

        services.append([
            item for sublist in data['services'].values() for item in sublist
        ])

    return set([item for sublist in services for item in sublist])


ALL_SERVICES = _get_all_services()

def service_table(format='simple'):
    """
    Returns a string depicting all services currently installed.
    """
    if format == 'html':
        linkify = lambda x: "<a href='{0}' target='_blank'>{0}</a>".format(x)
    else:
        linkify = lambda x: x

    ret = []
    for service in sorted(ALL_SERVICES, key=lambda x: x.service_id):
        ret.append([
            service.service_id,
            service.__name__, linkify(service.api_homepage.format(
                domain=service.domain, protocol=service.protocol
            )),
            ", ".join(service.supported_cryptos or [])
        ])

    return tabulate(ret, headers=['ID', 'Name', 'URL', 'Supported Currencies'], tablefmt=format)

def wif_to_hex(wif):
    """
    Convert a WIF encded private key and return the raw hex encoded private key
    This function works for all bitcoin-API compatable coins.
    """
    return hexlify(b58decode_check(wif)[1:]).upper()



from moneywagon import (
    get_unspent_outputs, CurrentPrice, get_optimal_fee, PushTx,
    get_onchain_exchange_rates
)
from moneywagon.core import get_optimal_services, get_magic_bytes
from bitcoin import mktx, sign, pubtoaddr, privtopub
from .crypto_data import crypto_data

class Transaction(object):
    def __init__(self, crypto, hex=None, verbose=False):
        if crypto.lower() in ['nxt']:
            raise NotImplementedError("%s not yet supported" % crypto.upper())

        self.change_address = None
        self.crypto = crypto
        self.fee_satoshi = None
        self.outs = []
        self.ins = []
        self.onchain_rate = None

        self.verbose = verbose

        services = get_optimal_services(self.crypto, 'current_price')
        self.price_getter = CurrentPrice(services=services, verbose=verbose)

        if hex:
            self.hex = hex

    def from_unit_to_satoshi(self, value, unit='satoshi'):
        """
        Convert a value to satoshis. units can be any fiat currency.
        By default the unit is satoshi.
        """
        if not unit or unit == 'satoshi':
            return value
        if unit == 'bitcoin' or unit == 'btc':
            return value * 1e8

        # assume fiat currency that we can convert
        convert = self.price_getter.action(self.crypto, unit)
        return int(value / convert * 1e8)

    def add_raw_inputs(self, inputs, private_key=None):
        """
        Add a set of utxo's to this transaction. This method is better to use if you
        want more fine control of which inputs get added to a transaction.
        `inputs` is a list of "unspent outputs" (they were 'outputs' to previous transactions,
          and 'inputs' to subsiquent transactions).

        `private_key` - All inputs will be signed by the passed in private key.
        """
        for i in inputs:
            self.ins.append(dict(input=i, private_key=private_key))
            self.change_address = i['address']

    def _get_utxos(self, address, services, **modes):
        """
        Using the service fallback engine, get utxos from remote service.
        """
        return get_unspent_outputs(
            self.crypto, address, services=services,
            **modes
        )

    def private_key_to_address(self, pk):
        """
        Convert a private key (in hex format) into an address.
        """
        pub = privtopub(pk)
        pub_byte, priv_byte = get_magic_bytes(self.crypto)

        if priv_byte >= 128:
            priv_byte -= 128 #pybitcointools bug

        return pubtoaddr(pub, pub_byte)

    def add_inputs(self, private_key=None, address=None, amount='all', max_ins=None, password=None, services=None, **modes):
        """
        Make call to external service to get inputs from an address and/or private_key.
        `amount` is the amount of [currency] worth of inputs (in satoshis) to add from
        this address. Pass in 'all' (the default) to use *all* inputs found for this address.
         Returned is the number of units (in satoshis) that were added as inputs to this tx.
        """
        if private_key:
            if private_key.startswith('6P'):
                if not password:
                    raise Exception("Password required for BIP38 encoded private keys")
                from .bip38 import Bip38EncryptedPrivateKey
                private_key = Bip38EncryptedPrivateKey(self.crypto, private_key).decrypt(password)

            address_from_priv = self.private_key_to_address(private_key)
            if address and address != address_from_priv:
                raise Exception("Invalid Private key")
            address = address_from_priv
            self.change_address = address

        if not services:
            services = get_optimal_services(self.crypto, 'unspent_outputs')

        total_added_satoshi = 0
        ins = 0
        for utxo in self._get_utxos(address, services, **modes):
            if max_ins and ins >= max_ins:
                break
            if (amount == 'all' or total_added_satoshi < amount):
                ins += 1
                self.ins.append(
                    dict(input=utxo, private_key=private_key)
                )
                total_added_satoshi += utxo['amount']

        return total_added_satoshi, ins

    def total_input_satoshis(self):
        """
        Add up all the satoshis coming from all input tx's.
        """
        just_inputs = [x['input'] for x in self.ins]
        return sum([x['amount'] for x in just_inputs])

    def select_inputs(self, amount):
      '''Maximize transaction priority. Select the oldest inputs,
      that are sufficient to cover the spent amount. Then,
      remove any unneeded inputs, starting with
      the smallest in value.
      Returns sum of amounts of inputs selected'''
      sorted_txin = sorted(self.ins, key=lambda x:-x['input']['confirmations'])
      total_amount = 0
      for (idx, tx_in) in enumerate(sorted_txin):
        total_amount += tx_in['input']['amount']
        if (total_amount >= amount):
          break
      sorted_txin = sorted(sorted_txin[:idx+1], key=lambda x:x['input']['amount'])
      for (idx, tx_in) in enumerate(sorted_txin):
        value = tx_in['input']['amount']
        if (total_amount - value < amount):
          break
        else:
          total_amount -= value
      self.ins = sorted_txin[idx:]
      return total_amount

    def add_output(self, address, value, unit='satoshi'):
        """
        Add an output (a person who will receive funds via this tx).
        If no unit is specified, satoshi is implied.
        """
        value_satoshi = self.from_unit_to_satoshi(value, unit)

        if self.verbose:
            print("Adding output of: %s satoshi (%.8f)" % (
                value_satoshi, (value_satoshi / 1e8)
            ))

        self.outs.append({
            'address': address,
            'value': value_satoshi
        })

    def onchain_exchange(self, withdraw_crypto, withdraw_address, value, unit='satoshi'):
        """
        This method is like `add_output` but it sends to another
        """
        self.onchain_rate = get_onchain_exchange_rates(
            self.crypto, withdraw_crypto, best=True, verbose=self.verbose
        )
        exchange_rate = float(self.onchain_rate['rate'])
        result = self.onchain_rate['service'].get_onchain_exchange_address(
            self.crypto, withdraw_crypto, withdraw_address
        )

        address = result['deposit']

        value_satoshi = self.from_unit_to_satoshi(value, unit)

        if self.verbose:
            print("Adding output of: %s satoshi (%.8f) via onchain exchange, converting to %s %s" % (
                value_satoshi, (value_satoshi / 1e8),
                exchange_rate * value_satoshi / 1e8, withdraw_crypto.upper()
            ))

        self.outs.append({
            'address': address,
            'value': value_satoshi
        })


    def fee(self, value=None, unit='satoshi'):
        """
        Set the miner fee, if unit is not set, assumes value is satoshi.
        If using 'optimal', make sure you have already added all outputs.
        """
        convert = None
        if not value:
            # no fee was specified, use $0.02 as default.
            convert = self.price_getter.action(self.crypto, "usd")
            self.fee_satoshi = int(0.02 / convert * 1e8)
            verbose = "Using default fee of:"

        elif value == 'optimal':
            self.fee_satoshi = get_optimal_fee(
                self.crypto, self.estimate_size(), verbose=self.verbose
            )
            verbose = "Using optimal fee of:"
        else:
            self.fee_satoshi = self.from_unit_to_satoshi(value, unit)
            verbose = "Using manually set fee of:"

        if self.verbose:
            if not convert:
                convert = self.price_getter.action(self.crypto, "usd")
            fee_dollar = convert * self.fee_satoshi / 1e8
            print(verbose + " %s satoshis ($%.2f)" % (self.fee_satoshi, fee_dollar))

    def estimate_size(self):
        """
        Estimate how many bytes this transaction will be by countng inputs
        and outputs.
        Formula taken from: http://bitcoin.stackexchange.com/a/3011/18150
        """
        # if there are no outs use 1 (because the change will be an out)
        outs = len(self.outs) or 1
        return outs * 34 + 148 * len(self.ins) + 10

    def get_hex(self, signed=True):
        """
        Given all the data the user has given so far, make the hex using pybitcointools
        """
        total_ins_satoshi = self.total_input_satoshis()
        if total_ins_satoshi == 0:
            raise ValueError("Can't make transaction, there are zero inputs")

        # Note: there can be zero outs (sweep or coalesc transactions)
        total_outs_satoshi = sum([x['value'] for x in self.outs])

        if not self.fee_satoshi:
            self.fee() # use default of $0.02

        change_satoshi = total_ins_satoshi - (total_outs_satoshi + self.fee_satoshi)

        if change_satoshi < 0:
            raise ValueError(
                "Input amount (%s) must be more than all output amounts (%s) plus fees (%s). You need more %s."
                % (total_ins_satoshi, total_outs_satoshi, self.fee_satoshi, self.crypto.upper())
            )

        ins = [x['input'] for x in self.ins]

        if change_satoshi > 0:
            if self.verbose:
                print("Adding change address of %s satoshis to %s" % (change_satoshi, self.change_address))
            change = [{'value': change_satoshi, 'address': self.change_address}]
        else:
            change = [] # no change ?!
            if self.verbose: print("Inputs == Outputs, no change address needed.")

        tx = mktx(ins, self.outs + change)

        if signed:
            for i, input_data in enumerate(self.ins):
                if not input_data['private_key']:
                    raise Exception("Can't sign transaction, missing private key for input %s" % i)
                tx = sign(tx, i, input_data['private_key'])

        return tx

    def push(self, services=None, redundancy=1):
        if not services:
            services = get_optimal_services(self.crypto, "push_tx")

        self.pushers = []
        pusher = PushTx(services=services, verbose=self.verbose)
        results = [pusher.action(self.crypto, self.get_hex())]

        try:
            for service in services[1:redundancy-1]:
                pusher = PushTx(services=[service], verbose=self.verbose)
                results.append(self.pusher.action(self.crypto, self.get_hex()))
                self.pushers.append(pusher)
        except:
            raise Exception("Partial push. Some services returned success, some failed.")

        return results



import json
from requests.auth import HTTPBasicAuth
from .core import Service, NoService, NoData, ServiceError, SkipThisService, currency_to_protocol
import arrow

class Bitstamp(Service):
    service_id = 1
    supported_cryptos = ['btc']
    api_homepage = "https://www.bitstamp.net/api/"
    name = "Bitstamp"

    def get_current_price(self, crypto, fiat):
        if fiat.lower() != 'usd':
            raise SkipThisService('Bitstamp only does USD->BTC')

        url = "https://www.bitstamp.net/api/ticker/"
        response = self.get_url(url).json()
        return float(response['last'])


class BlockCypher(Service):
    service_id = 2
    supported_cryptos = ['btc', 'ltc', 'doge']
    api_homepage = "http://dev.blockcypher.com/"

    explorer_address_url = "https://live.blockcypher.com/{crypto}/address/{address}"
    explorer_tx_url = "https://live.blockcypher.com/{crypto}/tx/{txid}"
    explorer_blockhash_url = "https://live.blockcypher.com/{crypto}/block/{blockhash}/"
    explorer_blocknum_url = "https://live.blockcypher.com/{crypto}/block/{blocknum}/"

    base_api_url = "https://api.blockcypher.com/v1/{crypto}"
    json_address_balance_url = base_api_url + "/main/addrs/{address}"
    json_txs_url = json_address_balance_url
    json_unspent_outputs_url = base_api_url + "/main/addrs/{address}/full?unspentOnly=true&includeScript=true"
    json_blockhash_url = base_api_url + "/main/blocks/{blockhash}"
    json_blocknum_url = base_api_url + "/main/blocks/{blocknum}"
    name = "BlockCypher"

    def get_balance(self, crypto, address, confirmations=1):
        url = self.json_address_balance_url.format(address=address, crypto=crypto)
        response = self.get_url(url)
        if confirmations == 0:
            return response.json()['final_balance'] / 1.0e8
        elif confirmations == 1:
            return response.json()['balance'] / 1.0e8
        else:
            raise SkipThisService("Filtering by confirmations only for 0 and 1")

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = self.json_unspent_outputs_url.format(address=address, crypto=crypto)
        utxos = []
        for utxo in self.get_url(url).json()['txs']:
            if utxo['confirmations'] < confirmations:
                continue
            for n, output in enumerate(utxo['outputs']):
                output_addrs = output['addresses']
                if address in output_addrs and len(output_addrs) == 1:
                    utxos.append(dict(
                        txid=utxo['hash'],
                        amount=output['value'],
                        output="%s:%s" % (utxo['hash'], n),
                        vout=n,
                        scriptPubKey=output['script'],
                        address=address,
                        confirmations=utxo['confirmations'],
                    ))
                elif address in output_addrs:
                    raise Exception("Multisig not implemented yet")

        return utxos

    def get_transactions(self, crypto, address, confirmations=1):
        url = self.json_txs_url.format(address=address, crypto=crypto)
        transactions = []
        for tx in self.get_url(url).json().get('txrefs', []):
            if tx['confirmations'] < confirmations:
                continue
            transactions.append(dict(
                date=arrow.get(tx['confirmed']).datetime,
                amount=tx['value'] / 1e8,
                txid=tx['tx_hash'],
                confirmations=tx['confirmations']
            ))
        return transactions

    def get_single_transaction(self, crypto, txid):
        url = "https://api.blockcypher.com/v1/%s/main/txs/%s" % (crypto, txid)
        tx = self.get_url(url).json()
        ins = [{'address': x['addresses'][0], 'amount': x['output_value']} for x in tx['inputs']]
        outs = [{'address': x['addresses'][0], 'amount': x['value']} for x in tx['outputs']]

        return dict(
            txid=txid,
            confirmations=tx['confirmations'],
            size=tx['size'],
            time=arrow.get(tx['received']).datetime,
            block_hash=tx.get('block_hash', None),
            block_number=tx['block_height'],
            inputs=ins,
            outputs=outs,
            total_in=sum(x['amount'] for x in ins),
            total_out=sum(x['amount'] for x in ins),
            fees=tx['fees'],
        )

    def get_optimal_fee(self, crypto, tx_bytes):
        url = "https://api.blockcypher.com/v1/%s/main" % crypto
        fee_kb = self.get_url(url).json()['high_fee_per_kb']
        return int(tx_bytes * fee_kb / 1024.0)


    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        if block_number == 0:
            raise SkipThisService("BlockCypher does not support block #0")

        if block_hash:
            url = self.json_blockhash_url.format(blockhash=block_hash, crypto=crypto)
        elif block_number != '':
            url = self.json_blocknum_url.format(blocknum=block_number, crypto=crypto)

        r = self.get_url(url).json()
        return dict(
            block_number=r['height'],
            confirmations=r['depth'] + 1,
            time=arrow.get(r['received_time']).datetime,
            sent_value=r['total'] / 1e8,
            total_fees=r['fees'] / 1e8,
            #mining_difficulty=r['bits'],
            hash=r['hash'],
            merkle_root=r['mrkl_root'],
            previous_hash=r['prev_block'],
            tx_count=r['n_tx'],
            txids=r['txids']
        )

class BlockSeer(Service):
    """
    This service has no publically documented API, this code was written
    from looking through chrome dev toolbar.
    """
    service_id = 3
    supported_cryptos = ['btc']
    api_homepage = "https://www.blockseer.com/about"

    explorer_address_url = "https://www.blockseer.com/addresses/{address}"
    explorer_tx_url = "https://www.blockseer.com/transactions/{txid}"
    explorer_blocknum_url = "https://www.blockseer.com/blocks/{blocknum}"
    explorer_blockhash_url = "https://www.blockseer.com/blocks/{blockhash}"

    json_address_balance_url = "https://www.blockseer.com/api/addresses/{address}"
    json_txs_url = "https://www.blockseer.com/api/addresses/{address}/transactions?filter=all"
    name = "BlockSeer"

    def check_error(self, response):
        if 'error' in response.json():
            raise SkipThisService("BlockSeer returned error: %s" % response.json()['error'])

    def get_balance(self, crypto, address, confirmations=1):
        url = self.json_address_balance_url.format(address=address)
        return self.get_url(url).json()['data']['balance'] / 1e8

    def get_transactions(self, crypo, address):
        url = self.json_txs_url.format(address=address)
        transactions = []
        for tx in self.get_url(url).json()['data']['address']['transactions']:
            transactions.append(dict(
                date=arrow.get(tx['time']).datetime,
                amount=tx['delta'] / 1e8,
                txid=tx['hash'],
            ))
        return transactions


class SmartBitAU(Service):
    service_id = 4
    api_homepage = "https://www.smartbit.com.au/api"
    base_url = "https://api.smartbit.com.au/v1/blockchain"
    explorer_address_url = "https://www.smartbit.com.au/address/{address}"
    explorer_tx_url = "https://www.smartbit.com.au/tx/{txid}"
    explorer_blocknum_url = "https://www.smartbit.com.au/block/{blocknum}"
    explorer_blockhash_url = "https://www.smartbit.com.au/block/{blockhash}"
    name = "SmartBit"

    supported_cryptos = ['btc']

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/address/%s" % (self.base_url, address)
        r = self.get_url(url).json()

        confirmed = float(r['address']['confirmed']['balance'])
        if confirmations > 1:
            return confirmed
        else:
            return confirmed + float(r['address']['unconfirmed']['balance'])

    def get_balance_multi(self, crypto, addresses, confirmations=1):
        url = "%s/address/%s" % (self.base_url, ",".join(addresses))
        response = self.get_url(url).json()

        ret = {}
        for data in response['addresses']:
            bal = float(data['confirmed']['balance'])
            if confirmations == 0:
                bal += float(data['unconfirmed']['balance'])
            ret[data['address']] = bal

        return ret

    def get_transactions(self, crypto, address, confirmations=1):
        url = "%s/address/%s" % (self.base_url, address)
        transactions = []
        for tx in self.get_url(url).json()['address']['transactions']:
            out_amount = sum(float(x['value']) for x in tx['outputs'] if address in x['addresses'])
            in_amount = sum(float(x['value']) for x in tx['inputs'] if address in x['addresses'])
            transactions.append(dict(
                amount=out_amount - in_amount,
                date=arrow.get(tx['time']).datetime,
                fee=float(tx['fee']),
                txid=tx['txid'],
                confirmations=tx['confirmations'],
            ))
        return transactions

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "%s/address/%s/unspent" % (self.base_url, address)
        utxos = []
        for utxo in self.get_url(url).json()['unspent']:
            utxos.append(dict(
                amount=utxo['value_int'],
                output="%s:%s" % (utxo['txid'], utxo['n']),
                address=address,
                txid=utxo['txid'],
                vout=utxo['n'],
                confirmations=utxo['confirmations'],
                scriptPubKey=utxo['script_pub_key']['hex'],
                scriptPubKey_asm=utxo['script_pub_key']['asm']
            ))
        return utxos

    def push_tx(self, crypto, tx_hex):
        """
        This method is untested.
        """
        url = "%s/pushtx" % self.base_url
        return self.post_url(url, {'hex': tx_hex}).content

    def get_mempool(self):
        url = "%s/transactions/unconfirmed?limit=1000" % self.base_url
        txs = []
        for tx in self.get_url(url).json()['transactions']:
            txs.append(dict(
                first_seen=arrow.get(tx['first_seen']).datetime,
                size=tx['size'],
                txid=tx['txid'],
                fee=float(tx['fee']),
            ))
        return txs


class Blockr(Service):
    service_id = 5
    supported_cryptos = ['btc', 'ltc', 'ppc', 'mec', 'qrk', 'dgc', 'tbtc']
    api_homepage = "http://blockr.io/documentation/api"

    explorer_address_url = "http://blockr.io/address/info/{address}"
    explorer_tx_url = "http://blockr.io/address/info/{txid}"
    explorer_blockhash_url = "http://blockr.io/block/info/{blockhash}"
    explorer_blocknum_url = "http://blockr.io/block/info/{blocknum}"
    explorer_latest_block = "http://blockr.io/block/info/latest"

    json_address_url = "http://{crypto}.blockr.io/api/v1/address/info/{address}"
    json_single_tx_url = "http://{crypto}.blockr.io/api/v1/tx/info/{txid}"
    json_txs_url = url = "http://{crypto}.blockr.io/api/v1/address/txs/{address}?unconfirmed=1"
    json_unspent_outputs_url = "http://{crypto}.blockr.io/api/v1/address/unspent/{address}"
    name = "Blockr.io"

    def get_balance(self, crypto, address, confirmations=1):
        url = self.json_address_url.format(address=address, crypto=crypto)
        response = self.get_url(url)
        return response.json()['data']['balance']

    def get_balance_multi(self, crypto, addresses, confirmations=1):
        url = self.json_address_url.format(address=','.join(addresses), crypto=crypto)
        balances = {}
        for bal in self.get_url(url).json()['data']:
            balances[bal['address']] = bal['balance']
        return balances

    def _format_tx(self, tx, address):
        return dict(
            date=arrow.get(tx['time_utc']).datetime,
            amount=tx['amount'],
            txid=tx['tx'],
            confirmations=tx['confirmations'],
            addresses=[address],
        )

    def get_transactions(self, crypto, address, confirmations=1):
        url = self.json_txs_url.format(address=address, crypto=crypto)
        response = self.get_url(url)

        transactions = []
        for tx in response.json()['data']['txs']:
            transactions.append(self._format_tx(tx, address))
        return transactions

    def get_transactions_multi(self, crypto, addresses, confirmation=1):
        url = self.json_txs_url.format(address=','.join(addresses), crypto=crypto)
        transactions = []
        for data in self.get_url(url).json()['data']:
            for tx in data['txs']:
                transactions.append(self._format_tx(tx, data['address']))
        return transactions

    def _format_single_tx(self, tx):
        ins = [{'address': x['address'], 'amount': currency_to_protocol(x['amount']) * -1} for x in tx['vins']]
        outs = [{'address': x['address'], 'amount': currency_to_protocol(x['amount'])} for x in tx['vouts']]

        return dict(
            time=arrow.get(tx['time_utc']).datetime,
            block_number=tx['block'],
            inputs=ins,
            outputs=outs,
            txid=tx['tx'],
            total_in=sum(x['amount'] for x in ins),
            total_out=sum(x['amount'] for x in outs),
            confirmations=tx['confirmations'],
            fee=float(tx['fee'])
        )

    def get_single_transaction(self, crypto, txid):
        url = self.json_single_tx_url.format(crypto=crypto, txid=txid)
        r = self.get_url(url).json()['data']
        return self._format_single_tx(r)


    def get_single_transaction_multi(self, crypto, txids):
        url = self.json_single_tx_url.format(crypto=crypto, txid=','.join(txids))
        txs = []
        for tx in self.get_url(url).json()['data']:
            txs.append(self._format_single_tx(tx))
        return txs

    def _format_utxo(self, utxo, address):
        return dict(
            amount=currency_to_protocol(utxo['amount']),
            address=address,
            output="%s:%s" % (utxo['tx'], utxo['n']),
            txid=utxo['tx'],
            vout=utxo['n'],
            confirmations=utxo['confirmations'],
            scriptPubKey=utxo['script']
        )

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = self.json_unspent_outputs_url.format(address=address, crypto=crypto)
        utxos = []
        for utxo in self.get_url(url).json()['data']['unspent']:
            cons = utxo['confirmations']
            if cons < confirmations:
                continue
            utxos.append(self._format_utxo(utxo, address))
        return utxos

    def get_unspent_outputs_multi(self, crypto, addresses, confirmations=1):
        url = self.json_unspent_outputs_url.format(address=','.join(addresses), crypto=crypto)
        utxos = []
        for data in self.get_url(url).json()['data']:
            for utxo in data['unspent']:
                cons = utxo['confirmations']
                if cons < confirmations:
                    continue
                utxos.append(self._format_utxo(utxo, data['address']))
        return utxos

    def push_tx(self, crypto, tx_hex):
        url = "http://%s.blockr.io/api/v1/tx/push" % crypto
        resp = self.post_url(url, {'tx': tx_hex}).json()
        if resp['status'] == 'fail':
            raise ValueError(
                "Blockr returned error: %s %s %s" % (
                    resp['code'], resp['data'], resp['message']
                )
            )
        return resp['data']

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        if block_number == 0:
            raise SkipThisService("Block 0 not supported (bug in Blockr.io?)")

        url ="http://%s.blockr.io/api/v1/block/info/%s%s%s" % (
            crypto,
            block_hash if block_hash != '' else '',
            block_number if block_number != '' else '',
            'last' if latest else ''
        )
        r = self.get_url(url).json()['data']
        return dict(
            block_number=r['nb'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time_utc']).datetime,
            sent_value=r['vout_sum'],
            total_fees=float(r['fee']),
            mining_difficulty=r['difficulty'],
            size=int(r['size']),
            hash=r['hash'],
            merkle_root=r['merkleroot'],
            previous_hash=r['prev_block_hash'],
            next_hash=r['next_block_hash'],
            tx_count=r['nb_txs'],
        )


class Toshi(Service):
    api_homepage = "https://toshi.io/docs/"
    service_id = 6
    url = "https://bitcoin.toshi.io/api/v0"
    name = "Toshi"

    supported_cryptos = ['btc']

    def check_error(self, response):
        if 'error' in response.json():
            raise SkipThisService("Toshi returned error: %s" % response.json()['error'])

        if response.status_code == 404:
            return # don't skip on 404

        super(Toshi, self).check_error(response)

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/addresses/%s" % (self.url, address)
        response = self.get_url(url).json()
        return response['balance'] / 1e8

    def get_transactions(self, crypto, address, confirmations=1):
        url = "%s/addresses/%s/transactions" % (self.url, address)
        response = self.get_url(url)
        if response.status_code == 404:
            return []

        j = response.json()
        transactions = []
        for tx in j.get('unconfirmed_transactions', []) + j['transactions']:
            transactions.append(dict(
                amount=sum([x['amount'] / 1e8 for x in tx['outputs'] if address in x['addresses']]),
                txid=tx['hash'],
                date=arrow.get(tx['block_time']).datetime if tx.get('block_time', False) else None,
                confirmations=tx['confirmations']
            ))

        return transactions

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "%s/addresses/%s/unspent_outputs" % (self.url, address)
        response = self.get_url(url).json()
        utxos = []
        for utxo in response:
            cons = utxo['confirmations']
            if cons < confirmations:
                continue
            utxos.append(dict(
                amount=utxo['amount'],
                address=address,
                output="%s:%s" % (utxo['transaction_hash'], utxo['output_index']),
                confirmations=cons,
                vout=utxo['output_index'],
                txid=utxo['transaction_hash'],
                scriptPubKey=utxo['script_hex'],
                scriptPubKey_asm=utxo['script']
            ))
        return utxos


    def push_tx(self, crypto, tx_hex):
        url = "%s/transactions/%s" % (self.url, tx_hex)
        return self.get_url(url).json()['hash']

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        if latest:
            url = "%s/blocks/latest" % self.url
        else:
            url = "%s/blocks/%s%s" % (
                self.url, block_hash if block_hash != '' else '',
                block_number if block_number != '' else ''
            )

        r = self.get_url(url).json()
        return dict(
            block_number=r['height'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time']).datetime,
            sent_value=r['total_out'] / 1e8,
            total_fees=r['fees'] / 1e8,
            mining_difficulty=r['difficulty'],
            size=r['size'],
            hash=r['hash'],
            merkle_root=r['merkle_root'],
            previous_hash=r['previous_block_hash'],
            next_hash=r['next_blocks'][0]['hash'] if len(r['next_blocks']) else None,
            txids=sorted(r['transaction_hashes']),
            tx_count=len(r['transaction_hashes'])
        )


class BTCE(Service):
    service_id = 7
    api_homepage = "https://btc-e.com/api/documentation"
    name = "BTCe"

    def get_current_price(self, crypto, fiat):
        pair = "%s_%s" % (crypto.lower(), fiat.lower())
        url = "https://btc-e.com/api/3/ticker/" + pair
        response = self.get_url(url).json()
        return response[pair]['last']


class Cryptonator(Service):
    service_id = 8
    api_homepage = "https://www.cryptonator.com/api"
    name = "Cryptonator"

    def get_current_price(self, crypto, fiat):
        pair = "%s-%s" % (crypto, fiat)
        url = "https://www.cryptonator.com/api/ticker/%s" % pair
        response = self.get_url(url).json()
        return float(response['ticker']['price'])


class Winkdex(Service):
    service_id = 9
    supported_cryptos = ['btc']
    api_homepage = "http://docs.winkdex.com/"
    name = "Winkdex"

    def get_current_price(self, crypto, fiat):
        if fiat != 'usd':
            raise SkipThisService("winkdex is btc->usd only")
        url = "https://winkdex.com/api/v0/price"
        return self.get_url(url).json()['price'] / 100.0,


class ChainSo(Service):
    service_id = 11
    api_homepage = "https://chain.so/api"
    base_url = "https://chain.so/api/v2"
    explorer_address_url = "https://chain.so/address/{crypto}/{address}"
    supported_cryptos = ['doge', 'btc', 'ltc']
    name = "Chain.So"

    def get_current_price(self, crypto, fiat):
        url = "%s/get_price/%s/%s" % (self.base_url, crypto, fiat)
        resp = self.get_url(url).json()
        items = resp['data']['prices']
        if len(items) == 0:
            raise SkipThisService("Chain.so can't get price for %s/%s" % (crypto, fiat))

        self.name = "%s via Chain.so" % items[0]['exchange']
        return float(items[0]['price'])

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/get_address_balance/%s/%s/%s" % (
            self.base_url, crypto, address, confirmations
        )
        response = self.get_url(url)
        return float(response.json()['data']['confirmed_balance'])

    def get_transactions(self, crypto, address, confirmations=1):
        url = "%s/get_tx_received/%s/%s" % (self.base_url, crypto, address)
        response = self.get_url(url)

        transactions = []
        for tx in response.json()['data']['txs']:
            tx_cons = int(tx['confirmations'])
            if tx_cons < confirmations:
                continue
            transactions.append(dict(
                date=arrow.get(tx['time']).datetime,
                amount=float(tx['value']),
                txid=tx['txid'],
                confirmations=tx_cons,
            ))

        return transactions

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "%s/get_tx_unspent/%s/%s" %(self.base_url, crypto, address)
        utxos = []
        for utxo in self.get_url(url).json()['data']['txs']:
            utxos.append(dict(
                amount=currency_to_protocol(utxo['value']),
                address=address,
                output="%s:%s" % (utxo['txid'], utxo['output_no']),
                confirmations=utxo['confirmations'],
                txid=utxo['txid'],
                vout=utxo['output_no'],
                scriptPubKey=utxo['script_hex'],
                scriptPubKey_asm=utxo['script_asm'],
            ))
        return utxos


    def push_tx(self, crypto, tx_hex):
        url = "%s/send_tx/%s" % (self.base_url, crypto)
        resp = self.post_url(url, {'tx_hex': tx_hex})
        return resp.json()['data']['txid']

    def get_block(self, crypto, block_number='', block_hash='', latest=False):
        if latest:
            raise SkipThisService("This service can't get block by latest")
        else:
            url = "%s/block/%s/%s%s" % (
                self.base_url, crypto, block_number, block_hash
            )
        r = self.get_url(url).json()['data']
        return dict(
            block_number=r['block_no'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time']).datetime,
            sent_value=float(r['sent_value']),
            total_fees=float(r['fee']),
            mining_difficulty=float(r['mining_difficulty']),
            size=r['size'],
            hash=r['blockhash'],
            merkle_root=r['merkleroot'],
            previous_hash=r['previous_blockhash'],
            next_hash=r['next_blockhash'],
            txids=sorted([t['txid'] for t in r['txs']])
        )


class CoinPrism(Service):
    service_id = 12
    api_homepage = "http://docs.coinprism.apiary.io/"
    base_url = "https://api.coinprism.com/v1"
    supported_cryptos = ['btc']
    name = "CoinPrism"

    def get_balance(self, crypto, address, confirmations=None):
        url = "%s/addresses/%s" % (self.base_url, address)
        resp = self.get_url(url).json()
        return resp['balance'] / 1e8

    def get_transactions(self, crypto, address):
        url = "%s/addresses/%s/transactions" % (self.base_url, address)
        transactions = []
        for tx in self.get_url(url).json():
            transactions.append(dict(
                amount=sum([x['value'] / 1e8 for x in tx['outputs'] if address in x['addresses']]),
                txid=tx['hash'],
                date=arrow.get(tx['block_time']).datetime,
                confirmations=tx['confirmations']
            ))

        return transactions

    def get_unspent_outputs(self, crypto, address):
        url = "%s/addresses/%s/unspents" % (self.base_url, address)
        transactions = []
        for tx in self.get_url(url).json():
            if address in tx['addresses']:
                transactions.append(dict(
                    amount=tx['value'],
                    address=address,
                    output="%s:%s" % (tx['transaction_hash'], tx['output_index']),
                    confirmations=tx['confirmations'],
                    txid=tx['transaction_hash'],
                    vout=tx['output_index'],
                    scriptPubKey=utxo['script_hex'],
                ))

        return transactions

    def push_tx(self, crypto, tx_hex):
        """
        Note: This one has not been tested yet.
        http://docs.coinprism.apiary.io/#reference/transaction-signing-and-broadcasting/push-a-signed-raw-transaction-to-the-network/post
        """
        url = "%s/sendrawtransaction"
        return self.post_url(url, tx_hex).content


class BitEasy(Service):
    """
    Most functions from this servie require an API key. therefore only
    address balance is supported at this time.
    """
    service_id = 13
    api_homepage = "https://support.biteasy.com/kb"
    supported_cryptos = ['btc']
    explorer_address_url = "https://www.biteasy.com/blockchain/addresses/{address}"
    explorer_tx_url = "https://www.biteasy.com/blockchain/transactions/{txid}"
    explorer_blockhash_url = "https://www.biteasy.com/blockchain/blocks/{blockhash}"
    name = "BitEasy"

    def check_error(self, response):
        if response.status_code == 404:
            msg = "; ".join(response.json()['messages'])
            raise SkipThisService("BitEasy returned 404: %s" % msg)

    def get_balance(self, crypto, address, confirmations=1):
        url = "https://api.biteasy.com/blockchain/v1/addresses/" + address
        response = self.get_url(url)
        return response.json()['data']['balance'] / 1e8


class BlockChainInfo(Service):
    service_id = 14
    domain = "blockchain.info"
    api_homepage = "https://{domain}/api"
    supported_cryptos = ['btc']
    explorer_address_url = "https://{domain}/address/{address}"
    explorer_tx_url = "https://{domain}/tx/{txid}"
    explorer_blocknum_url = "https://{domain}/block-index/{blocknum}"
    explorer_blockhash_url = "https://{domain}/block/{blockhash}"
    name = "Blockchain.info"

    def get_balance(self, crypto, address, confirmations=1):
        url = "https://%s/address/%s?format=json" % (self.domain, address)
        response = self.get_url(url)
        return float(response.json()['final_balance']) * 1e-8

    def get_single_transaction(self, crypto, txid):
        latest_block_number = self.get_block('btc', latest=True)['block_number']

        url = "https://%s/tx-index/%s?format=json" % (
            self.domain, txid
        )
        tx = self.get_url(url).json()
        outs = [{'address': x['addr'], 'amount': x['value']} for x in tx['out']]
        ins = []

        for in_ in tx['inputs']:
            if 'prev_out' in in_:
                prev = in_['prev_out']
                ins.append(
                    {'address': prev['addr'], 'amount': prev['value']}
                )

        block_number = tx.get('block_height', None)

        return dict(
            txid=txid,
            block_number=block_number,
            size=tx['size'],
            time=arrow.get(tx['time']).datetime,
            confirmations=(latest_block_number - block_number) if block_number else 0,
            inputs=ins,
            outputs=outs,
            total_in=sum(x['amount'] for x in ins),
            total_out=sum(x['amount'] for x in outs),
        )


    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "https://%s/unspent?active=%s" % (self.domain, address)

        response = self.get_url(url)
        if response.content == 'No free outputs to spend':
            return []

        utxos = []
        for utxo in response.json()['unspent_outputs']:
            if utxo['confirmations'] < confirmations:
                continue # don't return if too few confirmations

            utxos.append(dict(
                output="%s:%s" % (utxo['tx_hash_big_endian'], utxo['tx_output_n']),
                amount=utxo['value'],
                address=address,
                txid=utxo['tx_hash_big_endian'],
                vout=utxo['tx_output_n'],
                confirmations=utxo['confirmations'],
                scriptPubKey=utxo['script'],
            ))
        return utxos

    def get_block(self, crypto, block_number=None, block_hash=None, latest=False):
        if block_hash:
            raise SkipThisService("There is no way to get block by hash")

        url = "https://%s/latestblock" % self.domain
        latest_block_number = self.get_url(url).json()['height']

        if latest:
            block_number = latest_block_number

        url = "https://%s/block-height/%s?format=json" % (self.domain, block_number)
        r = self.get_url(url).json()['blocks'][0]
        confirmations = latest_block_number - r['height']
        return dict(
            block_number=r['height'],
            version=r['ver'],
            confirmations=confirmations,
            time=arrow.get(r['time']).datetime,
            size=r['size'],
            hash=r['hash'],
            merkle_root=r['mrkl_root'],
            previous_hash=r.get('prev_block', None),
            txids=[x['hash'] for x in r['tx']],
            tx_count=r['n_tx']
        )




##################################

class BitcoinAbe(Service):
    service_id = 15
    supported_cryptos = ['btc']
    base_url = "http://bitcoin-abe.info/chain/Bitcoin"
    name = "Abe"
    # decomissioned, kept here because other services need it as base class

    def get_balance(self, crypto, address, confirmations=1):
        url = self.base_url + "/q/addressbalance/" + address
        response = self.get_url(url)
        return float(response.content)


class DogeChainInfo(Service):
    service_id = 18
    supported_cryptos = ['doge']
    base_url = "https://dogechain.info/chain/Dogecoin"
    api_homepage = "https://dogechain.info/api/blockchain_api"
    name = "DogeChain.info"

    def get_balance(self, crypto, address, confirmations):
        url = "https://dogechain.info/api/v1/address/balance/" + address
        response = self.get_url(url).json()
        return response['balance']

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "https://dogechain.info/api/v1/unspent/" + address
        response = self.get_url(url).json()
        utxos = []
        for utxo in response['unspent_outputs']:
            utxos.append(dict(
                confirmations=utxo['confirmations'],
                amount=int(utxo['value']),
                scriptPubKey=utxo['script'],
                vout=utxo['tx_output_n'],
                txid=utxo['tx_hash'],
                output="%s:%s" % (utxo['tx_hash'], utxo['tx_output_n'])
            ))
        return utxos

    def get_single_transaction(self, crypto, txid):
        url = "https://dogechain.info/api/v1/transaction/" + txid
        tx = self.get_url(url).json()['transaction']
        return dict(
            txid=txid,
            confirmations=tx['confirmations'],
            size=tx['size'],
            time=arrow.get(r['time']).datetime,
            block_hash=r['block_hash'],
            inputs=[{'address': x['address'], 'amount': float(x['value'])} for x in r['inputs']],
            outputs=[{'address': x['address'], 'amount': float(x['value'])} for x in r['outputs']],
            total_in=r['total_input'],
            total_out=r['total_output'],
        )


class AuroraCoinEU(BitcoinAbe):
    service_id = 19
    supported_cryptos = ['aur']
    base_url = 'http://blockexplorer.auroracoin.eu/chain/AuroraCoin'
    name = "AuroraCoin.eu"


class Atorox(BitcoinAbe):
    service_id = 20
    supported_cryptos = ['aur']
    base_url = "http://auroraexplorer.atorox.net/chain/AuroraCoin"
    name = "atorox.net"

##################################

class FeathercoinCom(Service):
    service_id = 21
    supported_cryptos = ['ftc']
    api_homepage = "http://api.feathercoin.com/"
    name = "Feathercoin.com"

    def get_balance(self, crypto, address, confirmations=1):
        url= "http://api.feathercoin.com/?output=balance&address=%s&json=1" % address
        response = self.get_url(url)
        return float(response.json()['balance'])


class NXTPortal(Service):
    service_id = 22
    supported_cryptos = ['nxt']
    api_homepage = "https://nxtportal.org/"
    name = "NXT Portal"

    def get_balance(self, crypto, address, confirmations=1):
        url='http://nxtportal.org/nxt?requestType=getAccount&account=' + address
        response = self.get_url(url)
        return float(response.json()['balanceNQT']) * 1e-8

    def get_transactions(self, crypto, address):
        url = 'http://nxtportal.org/transactions/account/%s?num=50' % address
        response = self.get_url(url)
        transactions = []
        for tx in txs:
            transactions.append(dict(
                date=arrow.get(tx['time']).datetime,
                amount=tx['value'],
                txid=tx['txid'],
                confirmations=tx['confirmations'],
            ))

        return transactions


class CryptoID(Service):
    service_id = 23
    api_homepage = "https://chainz.cryptoid.info/api.dws"
    name = "CryptoID"
    api_key = "bc1063f00936"

    supported_cryptos = [
        'dash', 'bc', 'bay', 'block', 'cann', 'uno', 'vrc', 'xc', 'uro', 'aur',
        'pot', 'cure', 'arch', 'swift', 'karm', 'dgc', 'lxc', 'sync', 'byc',
        'pc', 'fibre', 'i0c', 'nobl', 'gsx', 'flt', 'ccn', 'rlc', 'rby', 'apex',
        'vior', 'ltcd', 'zeit', 'carbon', 'super', 'dis', 'ac', 'vdo', 'ioc',
        'xmg', 'cinni', 'crypt', 'excl', 'mne', 'seed', 'qslv', 'maryj', 'key',
        'oc', 'ktk', 'voot', 'glc', 'drkc', 'mue', 'gb', 'piggy', 'jbs', 'grs',
        'icg', 'rpc', 'tx'
    ]

    def get_balance(self, crypto, address, confirmations=1):
        url = "http://chainz.cryptoid.info/%s/api.dws?q=getbalance&a=%s&key=%s" % (
            crypto, address, self.api_key
        )
        return float(self.get_url(url).content)

    def get_single_transaction(self, crypto, txid):
        url = "http://chainz.cryptoid.info/%s/api.dws?q=txinfo&t=%s&key=%s" % (
            crypto, txid, self.api_key
        )
        r = self.get_url(url).json()

        return dict(
            time=arrow.get(r['timestamp']).datetime,
            block_number=r['block'],
            inputs=[{'address': x['addr'], 'amount': x['amount']} for x in r['inputs']],
            outputs=[{'address': x['addr'], 'amount': x['amount']} for x in r['outputs']],
            txid=txid,
            total_in=r['total_input'],
            total_out=r['total_output'],
            confirmations=r['confirmations'],
        )

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "http://chainz.cryptoid.info/%s/api.dws?q=unspent&active=%s&key=%s" % (
            crypto, address, self.api_key
        )

        resp = self.get_url(url)
        if resp.status_code != 200:
            raise Exception("CryptoID returned Error: %s" % resp.content)

        ret = []
        for utxo in resp.json()['unspent_outputs']:
            ret.append(dict(
                output="%s:%s" % (utxo['tx_hash'], utxo['tx_ouput_n']),
                amount=int(utxo['value']),
                confirmations=utxo['confirmations'],
                address=address,
                txid=utxo['tx_hash'],
                vout=utxo['tx_ouput_n'],
            ))
        return ret


class CryptapUS(Service):
    service_id = 24
    api_homepage = "https://cryptap.us/"
    name = "cryptap.us"
    supported_cryptos = [
        'nmc', 'wds', 'ber', 'scn', 'sc0', 'wdc', 'nvc', 'cas', 'myr'
    ]

    def get_balance(self, crypto, address, confirmations=1):
        url = "http://cryptap.us/%s/explorer/q/addressbalance/%s" % (crypto, address)
        return float(self.get_url(url).content)


class BTER(Service):
    service_id = 25
    api_homepage = "https://bter.com/api"
    name = "BTER"

    def get_current_price(self, crypto, fiat):
        url_template = "http://data.bter.com/api/1/ticker/%s_%s"
        url = url_template % (crypto, fiat)

        response = self.get_url(url).json()

        if response['result'] == 'false': # bter api returns this as string
            # bter doesn't support this pair, we need to make 2 calls and
            # do the math ourselves. The extra http request isn't a problem because
            # of caching. BTER only has USD, BTC and CNY
            # markets, so any other fiat will likely fail.

            url = url_template % (crypto, 'btc')
            response = self.get_url(url)
            altcoin_btc = float(response['last'])

            url = url_template % ('btc', fiat)
            response = self.get_url(url)
            btc_fiat = float(response['last'])

            self.name = 'BTER (calculated)'

            return (btc_fiat * altcoin_btc)

        return float(response['last'] or 0)

################################################

class BitpayInsight(Service):
    service_id = 28
    supported_cryptos = ['btc']
    domain = "insight.bitpay.com"
    protocol = 'https'
    api_homepage = "{protocol}://{domain}/api"
    explorer_address_url = "{protocol}://{domain}/address/{address}"
    api_tag = 'api'
    name = "Bitpay Insight"

    def check_error(self, response):
        if response.status_code == 400:
            raise ServiceError(response.content)

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s://%s/%s/addr/%s/balance" % (self.protocol, self.domain, self.api_tag, address)
        return float(self.get_url(url).content) / 1e8

    def _format_tx(self, tx, addresses):
        matched_addresses = []
        my_outs = 0
        my_ins = 0
        for address in addresses:
            for x in tx['vout']:
                if address in x['scriptPubKey']['addresses']:
                    my_outs += float(x['value'])
                    matched_addresses.append(address)
            for x in tx['vin']:
                if address in x['addr']:
                    my_ins += float(x['value'])
                    matched_addresses.append(address)

        return dict(
            amount=my_outs - my_ins,
            date=arrow.get(tx['time']).datetime if tx.get('time', False) else None,
            txid=tx['txid'],
            confirmations=tx.get('confirmations', 0),
            addresses=list(set(matched_addresses))
        )

    def get_transactions(self, crypto, address):
        url = "%s://%s/api/txs/?address=%s" % (self.protocol, self.domain, address)
        response = self.get_url(url)
        transactions = []
        for tx in response.json()['txs']:
            transactions.append(self._format_tx(tx, [address]))
        return transactions

    def get_transactions_multi(self, crypto, addresses):
        url = "%s://%s/api/addrs/%s/txs" % (self.protocol, self.domain, ','.join(addresses))
        r = self.get_url(url).json()
        txs = []
        for tx in r['items']:
            txs.append(self._format_tx(tx, addresses))
        return txs

    def get_single_transaction(self, crypto, txid):
        url = "%s://%s/api/tx/%s" % (self.protocol, self.domain, txid)
        d = self.get_url(url).json()

        block_time = None
        if d.get('blocktime'):
            block_time = arrow.get(d['blocktime']).datetime

        return dict(
            time=block_time,
            size=d['size'],
            confirmations=d['confirmations'] if block_time else 0,
            total_in=currency_to_protocol(d['valueIn']),
            total_out=currency_to_protocol(d['valueOut']),
            fee=currency_to_protocol(d['fees']),
            inputs=[{'address': x['addr'], 'amount': currency_to_protocol(x['value'])} for x in d['vin']],
            outputs=[{'address': x['scriptPubKey']['addresses'][0], 'amount': currency_to_protocol(x['value'])} for x in d['vout']],
            txid=txid,
        )

    def _format_utxo(self, utxo):
        return dict(
            txid=utxo['txid'],
            vout=utxo['vout'],
            scriptPubKey=utxo['scriptPubKey'],
            output="%s:%s" % (utxo['txid'], utxo['vout']),
            amount=currency_to_protocol(utxo['amount']),
            confirmations=utxo.get('confirmations', None),
            address=utxo['address']
        )

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "%s://%s/api/addr/%s/utxo?noCache=1" % (self.protocol, self.domain, address)
        utxos = []
        for utxo in self.get_url(url).json():
            utxos.append(self._format_utxo(utxo))
        return utxos

    def get_unspent_outputs_multi(self, crypto, addresses, confirmations=1):
        url = "%s://%s/api/addrs/%s/utxo?noCache=1" % (self.protocol, self.domain, ','.join(addresses))
        utxos = []
        for utxo in self.get_url(url).json():
            utxos.append(self._format_utxo(utxo))
        return utxos

    def get_block(self, crypto, block_number='', block_hash='', latest=False):

        if latest:
            url = "%s://%s/api/status?q=getLastBlockHash" % (self.protocol, self.domain)
            block_hash = self.get_url(url).json()['lastblockhash']

        elif block_number != '':
            url = "%s://%s/api/block-index/%s" % (self.protocol, self.domain, block_number)
            block_hash = self.get_url(url).json()['blockHash']

        url = "%s://%s/api/block/%s" % (self.protocol, self.domain, block_hash)

        r = self.get_url(url).json()
        return dict(
            block_number=r['height'],
            version=r['version'],
            confirmations=r['confirmations'],
            time=arrow.get(r['time']).datetime,
            mining_difficulty=float(r['difficulty']),
            size=r['size'],
            hash=r['hash'],
            merkle_root=r['merkleroot'],
            previous_hash=r.get('previousblockhash', None),
            next_hash=r.get('nextblockhash', None),
            txids=r['tx'],
            tx_count=len(r['tx'])
        )

    def push_tx(self, crypto, tx_hex):
        url = "%s://%s/api/tx/send" % (self.protocol, self.domain)
        return self.post_url(url, {'rawtx': tx_hex}).json()['txid']


    def get_optimal_fee(self, crypto, tx_bytes):
        url = "%s://%s/api/utils/estimatefee?nbBlocks=2" % (self.protocol, self.domain)
        return self.get_url(url).json()

class MYRCryptap(BitpayInsight):
    service_id = 30
    protocol = 'http'
    supported_cryptos = ['myr']
    domain = "insight-myr.cryptap.us"
    name = "MYR cryptap"


class BirdOnWheels(BitpayInsight):
    service_id = 31
    supported_cryptos = ['myr']
    domain = "birdonwheels5.no-ip.org:3000"
    name = "Bird on Wheels"


class Verters(BitpayInsight):
    service_id = 32
    supported_cryptos = ['vtc']
    domain = "explorer.verters.com"
    name = "Verters"


class ReddcoinCom(BitpayInsight):
    service_id = 33
    supported_cryptos = ['rdd']
    domain = "live.reddcoin.com"
    name = "Reddcoin.com"


class CoinTape(Service):
    service_id = 35
    api_homepage = "http://api.cointape.com/api"
    supported_cryptos = ['btc']
    base_url = "http://api.cointape.com"
    name = "CoinTape"

    def get_optimal_fee(self, crypto, tx_bytes):
        url = self.base_url + "/v1/fees/recommended"
        response = self.get_url(url).json()
        return int(response['fastestFee'] * tx_bytes)

class BitGo(Service):
    service_id = 36
    api_homepage = 'https://www.bitgo.com/api/'
    name = "BitGo"

    base_url = "https://www.bitgo.com"
    optimalFeeNumBlocks = 1
    supported_cryptos = ['btc']

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/api/v1/address/%s" % (self.base_url, address)
        response = self.get_url(url).json()
        if confirmations == 0:
            return response['balance'] / 1e8
        if confirmations == 1:
            return response['confirmedBalance'] / 1e8
        else:
            raise SkipThisService('Filtering by confirmation only available for 0 or 1')

    def get_transactions(self, crypto, address):
        url = "%s/api/v1/address/%s/tx" % (self.base_url, address)
        response = self.get_url(url).json()

        txs = []
        for tx in response['transactions']:
            my_outs = [x['value'] for x in tx['entries'] if x['account'] == address]

            txs.append(dict(
                amount=sum(my_outs),
                date=arrow.get(tx['date']).datetime,
                txid=tx['id'],
                confirmations=tx['confirmations'],
            ))
        return txs

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        url = "%s/api/v1/address/%s/unspents" % (self.base_url, address)
        utxos = []
        for utxo in self.get_url(url).json()['unspents']:
            utxos.append(dict(
                output="%s:%s" % (utxo['tx_hash'], utxo['tx_output_n']),
                amount=utxo['value'],
                confirmations=utxo['confirmations'],
                address=address,
                script=utxo['script'],
                vout=utxo['tx_output_n'],
            ))
        return utxos

    def get_block(self, crypto, block_number='', block_hash='', latest=False):
        if block_number == 0:
            raise SkipThisService("Block #0 broken for this service")

        if latest:
            url = "/api/v1/block/latest"
        elif block_number != '':
            url = "/api/v1/block/%s" % block_number
        else:
            url = "/api/v1/block/%s" % block_hash

        r = self.get_url(self.base_url + url).json()
        return dict(
            block_number=r['height'],
            time=arrow.get(r['date']).datetime,
            hash=r['id'],
            previous_hash=r['previous'],
            txids=r['transactions'],
            tx_count=len(r['transactions'])
        )

    def get_optimal_fee(self, crypto, tx_bytes):
        url = "%s/api/v1/tx/fee?numBlocks=%s" % (self.base_url, self.optimalFeeNumBlocks)
        response = self.get_url(url).json()
        fee_kb = response['feePerKb']
        return int(tx_bytes * fee_kb / 1024)

class Blockonomics(Service):
    service_id = 37
    supported_cryptos = ['btc']
    api_homepage = "https://www.blockonomics.co/views/api.html"
    name = "Blockonomics"
    base_url = "https://www.blockonomics.co"

    def get_balance(self, crypto, address, confirmations=1):
        return self.get_balance_multi(crypto, [address], confirmations)[address]

    def get_balance_multi(self, crypto, addresses, confirmations=1):
        url = "%s/api/balance" % self.base_url

        if hasattr(addresses, 'startswith') and addresses.startswith("xpub"):
            body = {'addr': addresses}
        else:
            body = {'addr': ' '.join(addresses)}

        response = self.post_url(url, json.dumps(body)).json()
        balances = {}
        for data in response['response']:
            confirmed = data['confirmed'] / 1e8
            if confirmations == 0:
                balance = confirmed + (data['unconfirmed'] / 1e8)
            if confirmations == 1:
                balance = confirmed
            else:
                raise SkipThisService("Can't filter by confirmations")

            balances[data['addr']] = balance

        return balances

    def get_transactions(self, crypto, address):
        url = "%s/api/searchhistory" % self.base_url
        response = self.post_url(url, json.dumps({'addr': address})).json()
        txs = []
        for tx in response['history'] + response.get('pending', []):
            txs.append(dict(
                amount=tx['value'] / 1e8,
                date=arrow.get(tx['time']).datetime,
                txid=tx['txid'],
            ))
        return txs

    def get_single_transaction(self, crypto, txid):
        url = "%s/api/tx_detail?txid=%s" % (self.base_url, txid)
        d = self.get_url(url).json()
        return dict(
            confirmations=1 if d['status'] == 'Confirmed' else 0,
            time=arrow.get(d['time']).datetime,
            inputs=[{'address': x['address'], 'amount': x['value']} for x in d['vin']],
            outputs=[{'address': x['address'], 'amount': x['value']} for x in d['vout']],
            txid=txid,
            fees=d['fee'],
            size=d['size']
        )


class BlockExplorerCom(BitpayInsight):
    service_id = 38
    supported_cryptos = ['btc']
    domain = "blockexplorer.com"
    name = "BlockExplorer.com"

class BitNodes(Service):
    domain = "https://bitnodes.21.co"
    service_id = 39
    name = "Bitnodes.21.co"

    def get_nodes(self, crypto):
        response = self.get_url(self.domain + "/api/v1/snapshots/latest/")
        nodes_dict = response.json()['nodes']

        nodes = []
        for address, data in nodes_dict.items():
            nodes.append({
                'address': address,
                'protocol_version': data[0],
                'user_agent': data[1],
                'connected_since': arrow.get(data[2]).datetime,
                'services': data[3],
                'height': data[4],
                'hostname': data[5],
                'city': data[6],
                'country': data[7],
                'latitude': data[8],
                'longitude': data[9],
                'timezone': data[10],
                'asn': data[11],
                'organization': data[12]
            })

        return nodes

class BitcoinFees21(CoinTape):
    base_url = "https://bitcoinfees.21.co/api"
    service_id = 40
    name = "bitcoinfees.21.co"
    api_homepage = "https://bitcoinfees.21.co/api"
    supported_cryptos = ['btc']


class ChainRadar(Service):
    api_homepage = "http://chainradar.com/api"
    service_id = 41
    name = "ChainRadar.com"
    supported_cryptos = ['aeon', 'bbr', 'bcn', 'btc', 'dsh', 'fcn', 'mcn', 'qcn', 'duck', 'mro', 'rd']

    def get_block(self, crypto, block_number='', block_hash='', latest=False):
        if latest:
            url = "http://chainradar.com/api/v1/%s/status" % crypto
            block_number = self.get_url(url).json()['height']

        url = "http://chainradar.com/api/v1/%s/blocks/%s/full" % (crypto, block_number or block_hash)
        r = self.get_url(url).json()
        h = r['blockHeader']

        return dict(
            block_number=h['height'],
            time=arrow.get(h['timestamp']).datetime,
            size=h['blockSize'],
            hash=h['hash'],
            previous_hash=h['prevBlockHash'],
            txids=[x['hash'] for x in r['transactions']],
            tx_count=len(r['transactions'])
        )

class Mintr(Service):
    service_id = 42
    name = "Mintr.org"
    domain = "http://{coin}.mintr.org"
    supported_cryptos = ['ppc', 'emc']
    api_homepage = "https://www.peercointalk.org/index.php?topic=3998.0"
    explorer_tx_url = "https://{coin}.mintr.org/tx/{txid}"
    explorer_address_url = "https://{coin}.mintr.org/address/{address}"
    explorer_blocknum_url = "https://{coin}.mintr.org/block/{blocknum}"
    explorer_blockhash_url = "https://{coin}.mintr.org/block/{blockhash}"

    @classmethod
    def _get_coin(cls, crypto):
        if crypto == 'ppc':
            return 'peercoin'
        if crypto == 'emc':
            return 'emercoin'

    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/api/address/balance/%s" % (
            self.domain.format(coin=self._get_coin(crypto)), address
        )
        r = self.get_url(url).json()

        if 'error' in r:
            raise Exception("Mintr returned error: %s" % r['error'])

        return float(r['balance'])

    def get_transactions(self, crypto, address):
        url = "%s/api/address/balance/%s/full" % (
            self.domain.format(coin=self._get_coin(crypto)), address
        )
        txs = []
        for tx in self.get_url(url).json()['transactions']:
            if not tx['sent'] and not tx['received']:
                continue

            amount = float(tx['sent'] or tx['received'])
            txs.append(dict(
                address=address,
                amount=amount if tx['received'] else -1 * amount,
                date=arrow.get(tx['time']).datetime,
                txid=tx['tx_hash'],
            ))
        return txs

    def get_single_transaction(self, crypto, txid):
        url = "%s/api/tx/hash/%s" % (
            self.domain.format(coin=self._get_coin(crypto)), txid
        )

        d = self.get_url(url).json()
        return dict(
            time=arrow.get(d['time']).datetime,
            total_in=float(d['valuein']),
            total_out=float(d['valueout']),
            fee=float(d['fee']),
            inputs=[{'address': x['address'], 'amount': x['value']} for x in d['vin']],
            outputs=[{'address': x['address'], 'amount': x['value']} for x in d['vout']],
            txid=txid,
        )

    def get_block(self, crypto, block_number='', block_hash='', latest=False):
        by = "latest"
        if block_number:
            by = "height/" + block_number
        elif block_hash:
            by = "hash/" + block_hash

        url = "%s/api/block/%s" % (
            self.domain.format(coin=self._get_coin(crypto)), by
        )

        b = self.get_url(url).json()

        return dict(
            block_number=int(b['height']),
            time=arrow.get(b['time']).datetime,
            hash=b['blockhash'],
            previous_hash=b['previousblockhash'],
            txids=[x['tx_hash'] for x in b['transactions']],
            tx_count=int(b['numtx']),
            size=int(b['size']),
            sent_value=float(b['valueout']) + float(b['mint']),
            mining_difficulty=float(b['difficulty']),
            merkle_root=b['merkleroot'],
            total_fees=float(b['fee'])
        )


class BlockExplorersNet(Service):
    service_id = 43
    domain = "http://{coin}.blockexplorers.net"
    supported_cryptos = ['gsm', 'erc', 'tx']
    name = "BlockExplorers.Net"

    explorer_tx_url = "https://{coin}.blockexplorers.net/tx/{txid}"
    explorer_address_url = "https://{coin}.blockexplorers.net/address/{address}"
    #explorer_blocknum_url = "https://{coin}.blockexplorers.net/block/{blocknum}"
    explorer_blockhash_url = "https://{coin}.blockexplorers.net/block/{blockhash}"

    @classmethod
    def _get_coin(cls, crypto):
        if crypto == 'gsm':
            return "gsmcoin"
        if crypto == 'erc':
            return 'europecoin'
        if crypto == 'tx':
            return 'transfercoin'


    def get_balance(self, crypto, address, confirmations=1):
        url = "%s/ext/getbalance/%s" % (
            self.domain.format(coin=self._get_coin(crypto)), address
        )
        return float(self.get_url(url).content)

    def get_transactions(self, crypto, address):
        domain = self.domain.format(coin=self._get_coin(crypto))
        url = "%s/ext/getaddress/%s" % (domain, address)
        return self.get_url(url).json()

    def get_single_transaction(self, crypto, txid):
        domain = self.domain.format(coin=self._get_coin(crypto))
        url = "%s/api/getrawtransaction?txid=%s&decrypt=1" % (domain, txid)

        d = self.get_url(url).json()

        if not d['vin'] or not d['vin'][0].get('coinbase'):
            ins = [{'txid': x['txid']} for x in d['vin']]
        else:
            ins = [{'txid': x['coinbase']} for x in d['vin']]

        outs = [{'address': x['scriptPubKey']['addresses'][0], 'amount': x['value']} for x in d['vout']]

        return dict(
            time=arrow.get(d['time']).datetime,
            block_hash=d['blockhash'],
            hex=d['hex'],
            inputs=ins,
            outputs=outs,
            txid=txid,
            total_out=sum(x['amount'] for x in outs),
            confirmations=d['confirmations'],
        )

    def get_block(self, crypto, block_number='', block_hash='', latest=False):
        domain = self.domain.format(coin=self._get_coin(crypto))

        if latest:
            url = "%s/api/getblockcount" % domain
            block_number = int(self.get_url(url).content)

        if block_number:
            url = "%s/api/getblockhash?index=%s" % (domain, block_number)
            block_hash = self.get_url(url).content

        url = "%s/api/getblock?hash=%s" % (domain, block_hash)
        r = self.get_url(url).json()

        return dict(
            confirmations=r['confirmations'],
            size=r['size'],
            txs=r['tx'],
            tx_count=len(r['txs']),
            time=arrow.get(r['time']).datetime,
            hash=r['hash'],
            block_number=r['height'],
            merkle_root=r['merkleroot'],
            difficulty=r['difficulty'],
        )

class UNOCryptap(BitpayInsight):
    service_id = 44
    supported_cryptos = ['uno']
    protocol = 'http'
    domain = "insight-uno.cryptap.us"
    name = "UNO cryptap"

class RICCryptap(BitpayInsight):
    service_id = 45
    supported_cryptos = ['ric']
    protocol = 'http'
    domain = "insight-ric.cryptap.us"
    name = "RIC cryptap"


class ProHashing(Service):
    service_id = 46
    domain = "prohashing.com"
    name= "ProHashing"

    def get_balance(self, crypto, address, confirmations=1):
        url = "https://%s/explorerJson/getAddress?address=%s&coin_id=%s" % (
            self.domain, address, self._get_coin(crypto)
        )
        r = self.get_url(url).json()

        if r.get('message', None):
            raise SkipThisService("Could not get Coin Info: %s" % r['message'])

        return r['balance']

    def get_transactions(self, crypto, address, confirmations=1):
        params = '$params={"page":1,"count":20,"filter":{},"sorting":{"blocktime":"asc"},"group":{},"groupBy":null}'
        url = "https://%s/explorerJson/getTransactionsByAddress?%s&address=%s&coin_id=%s" % (
            self.domain, params, address, self._get_coin(crypto)
        )
        txs = []
        for tx in self.get_url(url).json()['data']:
            txs.append(dict(
                amount=tx['value'],
                date=arrow.get(tx['blocktime']).datetime,
                txid=tx['transaction_hash'],
            ))
        return txs

    def get_single_transaction(self, crypto, txid):
        url = "https://%s/explorerJson/getTransaction?coin_id=%s&transaction_id=%s" % (
            self.domain, self._get_coin(crypto), txid
        )
        tx = self.get_url(url).json()

        return dict(
            time=arrow.get(tx['blocktime'] / 1000).datetime,
            block_hash=tx['block_hash'],
            block_number=tx['block_height'],
            inputs=[dict(address=x['address'], amount=-x['value']) for x in tx['address_inputs']],
            outputs=[dict(address=x['address'], amount=x['value']) for x in tx['address_outputs']],
            txid=txid,
            total_in=tx['total_outputs'],
            total_out=tx['total_outputs'],
            confirmations=tx['confirmations'],
        )

    def _get_coin(self, crypto):
        from crypto_data import crypto_data
        full_name = crypto_data[crypto]['name']
        url = "https://%s/explorerJson/getInfo?coin_name=%s" % (
            self.domain, full_name
        )
        r = self.get_url(url).json()

        if r.get('message', None):
            raise SkipThisService("Could not get Coin Info: %s" % r['message'])

        return r['id']

    def get_block(self, crypto, block_number='', block_hash='', latest=False):
        if latest or block_number:
            raise SkipThisService("Block height and latest not implemented.")

        url = "https://%s/explorerJson/getBlock?coin_id=%s&"% (
            self.domain, self._get_coin(crypto),
        )
        if block_hash:
            url = "%shash=%s" % (
                url, block_hash
            )

        r = self.get_url(url).json()
        return dict(
            confirmations=r['confirmations'],
            size=r['size'],
            txs=[x['hash'] for x in r['tx']],
            tx_count=len(r['tx']),
            time=arrow.get(r['time'] / 1000).datetime,
            hash=r['hash'],
            block_number=r['height'],
            difficulty=r['difficulty'],
        )

class SiampmDashInsight(BitpayInsight):
    service_id = 47
    supported_cryptos = ['dash']
    protocol = "http"
    domain = "insight.dash.siampm.com"
    name = "Siampm Dash Insight"

class BlockExperts(Service):
    service_id = 48
    name = "Block Experts"
    base = "https://www.blockexperts.com"
    supported_cryptos = ['hemp', 'dime', 'dope']

    def _get_coin(self, crypto):
        if crypto == 'hemp':
            return 33
        if crypto == 'dime':
            return 57
        if crypto == 'dope':
            return 19

    def get_address_balance(self, crypto, address, confirmations=1):
        url = "%s/api?coin=%s&action=getbalance&address=%s" % (
            self.base, crypto, address
        )
        return float(self.get_url(url).content)

    def get_single_transaction(self, crypto, txid):
        url = "%s/include/ajax/ajax.tx.raw.php?coin_id=%s&tx=%s" % (
            self.base, self._get_coin(crypto), txid
        )
        resp = self.get_url(url).json()
        return dict(
            time=arrow.get(tx['blocktime']).datetime,
            block_hash=tx['blockhash'],
            inputs=[dict(address=x['address'], amount=-x['value']) for x in tx['vin']],
            outputs=[dict(address=x['scriptPubKey']['address'], amount=x['value']) for x in tx['vout']],
            txid=txid,
            version=tx['version'],
            confirmations=tx['confirmations'],
        )

    def get_block(self, crypto, block_number=None, block_hash=None, latest=False):
        if latest:
            raise SkipThisService("Cant get by latest")
            url = "https://www.blockexperts.com/api?coin=hemp&action=getheight"
            block_height = int(self.get_url(url).content)

        if block_number is not None:
            raise SkipThisService("Cant get by block number")

        url = "%s/include/ajax/ajax.block.raw.php?coin_id=%s&block_hash=%s" % (
            self.base, self._get_coin(crypto), block_hash
        )
        r = self.get_url(url).json()
        return dict(
            confirmations=r['confirmations'],
            size=r['size'],
            txs=r['tx'],
            tx_count=len(r['tx']),
            time=arrow.get(r['time']).datetime,
            hash=r['hash'],
            block_number=r['height'],
            difficulty=r['difficulty'],
            merkle_root=r['merkleroot'],
        )

class MultiCoins(Service):
    service_id = 49
    supported_cryptos = ['ppc']

    def push_tx(self, crypto, tx_hex):
        """
        This method is untested.
        """
        url = "https://multicoins.org/api/v1/tx/push/ppc"
        return self.post_url(url, {'hex': tx_hex})

class BitcoinChain(Service):
    service_id = 50
    base = "https://api-r.bitcoinchain.com"
    supported_cryptos = ['btc']

    def get_address_balance(self, crypto, address, confirmations=1):
        url = "%s/v1/address/%s" % (self.base, address)
        return self.get_url(url).json()[0]['balance']

    def get_address_balance_multi(self, crypto, addresses, confirmations=1):
        url = "%s/v1/address/%s" % (self.base, ','.join(addresses))
        ret = {}
        for address_data in self.get_url(url).json():
            address = address_data['address']
            ret[address] = address_data['balance']

        return ret

    def get_transactions(self, crypto, address, confirmations=1):
        url = "%s/v1/address/txs/%s" % (self.base, address)
        response = self.get_url(url).json()[0]
        txs = []
        for tx in response:
            txs.append(dict(
                txid=tx['tx']['self_hash'],
                date=arrow.get(tx['tx']['block_time']).datetime,
            ))
        return txs

    def get_single_transaction(self, crypto, txid):
        url = "%s/v1/tx/%s" % (self.base, txid)
        r = self.get_url(url).json()

class NeoCrypto(BitpayInsight):
    service_id = 51
    protocol = "https"
    domain = "insight.neocrypto.io"
    api_tag = 'insight-api'

class CounterParty(Service):
    protocol = "http"
    port = None
    username = None
    password = None

    def check_error(self, response):
        j = response.json()
        if 'error' in j:
            e = j['error']
            raise Exception("Error code: %s %s" % (e['code'], e['message']))

        super(CounterParty, self).check_error(response)

    def authed_post_url(self, payload):
        auth = HTTPBasicAuth(self.username, self.password)
        headers = {'content-type': 'application/json'}
        url = "%s://%s:%s/api/" % (self.protocol, self.domain, self.port)
        return self.post_url(url, data=json.dumps(payload), auth=auth, headers=headers).json()

    def get_balance(self, crypto, address, confirmations=1):
        payload = {
            "method": "get_balances",
            "params": {
                "filters": [{"field": "address", "op": "==", "value": address}]
            },
            "jsonrpc": "2.0",
            "id": 0,
        }
        results = self.authed_post_url(payload)['result']
        for r in results:
            if r['address'] == address and r['asset'].lower() == crypto.lower():
                return r['quantity'] / 1e8

    def get_balance_multi(self, crypto, addresses, confirmations=1):
        payload = {
            "method": "get_balances",
            "params": {
                "filters": [{"field": "address", "op": "==", "value": x} for x in addresses],
                "filterop": "or"
            },
            "jsonrpc": "2.0",
            "id": 0,
        }
        results = self.authed_post_url(payload)['result']
        ret = {}
        for r in results:
            ret[r['address']] = r['quantity'] / 1e8

        return ret

    def _format_txs(self, credits, debits):
        txs = []
        for i, results in enumerate([credits, debits]):
            if i == 0:
                sign = 1
            else:
                sign = -1

            for tx in results:
                if tx['asset'].upper() != crypto.upper():
                    continue
                txs.append(dict(
                    amount=tx['quantity'] / 1e8 * sign,
                    txid=tx['event'],
                    address=tx['address'],
                    date=None,
                    counterparty=True
                ))
        return txs

    def get_transactions(self, crypto, address, confirmations=1):
        payload = {
            "method": "get_credits",
            "params": {
                "filters": [
                    {"field": "address", "op": "==", "value": address},
                    {"field": "asset", "op": "==", "value": crypto.upper()},
                ]
            },
            "jsonrpc": "2.0",
            "id": 0,
        }
        credits = self.authed_post_url(payload)['result']

        payload = {
            "method": "get_debits",
            "params": {
                "filters": [
                    {"field": "address", "op": "==", "value": address},
                    {"field": "asset", "op": "==", "value": crypto.upper()},
                ]
            },
            "jsonrpc": "2.0",
            "id": 0,
        }
        debits = self.authed_post_url(payload)['result']
        return self._format_txs(credits, debits)

    def get_transactions_multi(self, crypto, addresses):
        payload = {
            "method": "get_credits",
            "params": {
                "filters": [
                    {"field": "address", "op": "==", "value": address} for address in addresses
                ],
                "filterop": "or"
            },
            "jsonrpc": "2.0",
            "id": 0,
        }
        credits = self.authed_post_url(payload)['result']

        payload = {
            "method": "get_debits",
            "params": {
                "filters": [
                    {"field": "address", "op": "==", "value": address} for address in addresses
                ],
                "filterop": "or"
            },
            "jsonrpc": "2.0",
            "id": 0,
        }
        debits = self.authed_post_url(payload)['result']
        return self._format_txs(credits, debits)

    def get_single_transaction(self, crypto, txid):
        from moneywagon import get_single_transaction
        tx = get_single_transaction('btc', txid)
        return tx

    def make_unsigned_move_tx(self, crypto, amount, from_address, to_address):
        payload = {
            "method": "create_send",
            "params": {
                "source": from_address,
                "destination": to_address,
                "asset": crypto.upper(),
                "quantity": int(amount * 1e8),
                'encoding': "opreturn"
            },
            "jsonrpc": "2.0",
            "id": 0,
        }
        return self.authed_post_url(payload)['result']

    def get_unspent_outputs(self, crypto, address, confirmations=1):
        raise Exception("CounterParty does not use unspent outputs")

    def push_tx(self, crypto, tx_hex):
        from moneywagon import push_tx
        return push_tx('btc', tx_hex, random=True)

class CoinDaddy1(CounterParty):
    service_id = 52
    port = 4000
    domain = "public.coindaddy.io"
    username = 'rpc'
    password = '1234'
    name = "Coin Daddy #1"

class CoinDaddy2(CoinDaddy1):
    service_id = 53
    port = 4100
    name = "Coin Daddy #2"

class CounterPartyChain(Service):
    service_id = 54
    api_homepage = "https://counterpartychain.io/api"

    def get_balance(self, crypto, address, confirmations=1):
        url = "https://counterpartychain.io/api/balances/%s" % address
        response = self.get_url(url).json()
        if response['error']:
            return 0

        for balance in response['data']:
            if balance['asset'].upper() == crypto.upper():
                return float(balance['amount'])

    def push_tx(self, crypto, tx_hex):
        from moneywagon import push_tx
        return push_tx('btc', tx_hex, random=True)

class EtherChain(Service):
    service_id = 55
    api_homepage = "https://etherchain.org/documentation/api/"

    def get_current_price(self, crypto, fiat):
        url = "https://etherchain.org/api/basic_stats"
        price = self.get_url(url).json()['data']['price']
        if fiat.lower() in ['btc', 'usd']:
            return price[fiat.lower()]
        return self.convert_currency('usd', price['usd'], fiat)

    def get_balance(self, crypto, address, confirmations=1):
        url = "https://etherchain.org/api/account/%s" % address
        data = self.get_url(url).json()['data']
        return data[0]['balance'] / 1e18



from __future__ import print_function
import random
import requests
import time

from concurrent import futures

useragent = 'moneywagon 1.11.2'

class ServiceDisagreement(Exception):
    pass

class NoService(Exception):
    pass

class SkipThisService(NoService):
    pass

class ServiceError(NoService):
    pass

class NoData(Exception):
    pass

class RevertToPrivateMode(NotImplementedError):
    pass

class Service(object):
    """
    Represents a blockchain service running an Http interface.
    Some `Services` subclass will only support a subset of all pissible blockchain functions.
    All Services should subclass this class, and implement their own `get_*` method.
    """
    domain = None # optional, useful if the service will have subclasses
    api_homepage = '' # link to page defining the API.
    name = ''
    protocol = 'https'
    supported_cryptos = None # must be a list of lower case currency codes.
    explorer_address_url = None # url to block explerer page. Use {address} and {crypto} as placeholders.
    explorer_tx_url = None # {txid}
    explorer_blocknum_url = None # {blocknum}
    explorer_blockhash_url = None # {blockhash}

    def __init__(self, verbose=False, responses=None, timeout=None, random_wait_seconds=0):
        self.responses = responses or {} # for caching
        self.verbose = verbose
        self.last_url = None
        self.last_raw_response = None
        self.timeout = timeout
        self.random_wait_seconds = random_wait_seconds

    def __repr__(self):
        return "<Service: %s (%s in cache)>" % (self.__class__.__name__, len(self.responses))

    def get_url(self, url, *args, **kwargs):
        return self._external_request('get', url, *args, **kwargs)

    def post_url(self, url, *args, **kwargs):
        return self._external_request('post', url, *args, **kwargs)

    def check_error(self, response):
        """
        If the service is returning an error, this function should raise an exception.
        such as SkipThisService
        """
        if response.status_code == 503:
            raise SkipThisService("503 - Temporarily out of service.")

        if response.status_code == 429:
            raise SkipThisService("429 - Too many requests")

        if response.status_code == 404:
            raise SkipThisService("404 - Not Found")

    def convert_currency(self, base_fiat, base_amount, target_fiat):
        """
        Convert one fiat amount to another fiat. Uses the fixer.io service.
        """
        url = "http://api.fixer.io/latest?base=%s" % base_fiat
        data = self.get_url(url).json()
        try:
            return data['rates'][target_fiat.upper()] * base_amount
        except KeyError:
            raise Exception("Can not convert %s to %s" % (base_fiat, target_fiat))

    def _external_request(self, method, url, *args, **kwargs):
        """
        Wrapper for requests.get with useragent automatically set.
        And also all requests are reponses are cached.
        """
        self.last_url = url
        if url in self.responses.keys() and method == 'get':
            return self.responses[url] # return from cache if its there

        headers = kwargs.pop('headers', None)
        custom = {'User-Agent': useragent}
        if headers:
            headers.update(custom)
            kwargs['headers'] = headers
        else:
            kwargs['headers'] = custom

        if self.timeout:
            # add timeout parameter to requests.get if one was passed in on construction...
            kwargs['timeout'] = self.timeout

        response = getattr(requests, method)(url, *args, **kwargs)

        if self.verbose:
            print("Got Response: %s" % url)

        self.last_raw_response = response

        self.check_error(response)

        if method == 'get':
            self.responses[url] = response # cache for later

        return response

    def check_error(self, response):
        return

    def get_current_price(self, crypto, fiat):
        """
        Makes call to external service, and returns the price for given
        fiat/crypto pair. Returned is a float.
        """
        raise NotImplementedError(
            "This service does not support getting the current fiat exchange rate. "
            "Or rather it has no defined 'get_current_price' method."
        )

    def get_historical_price(self, crypto, fiat, at_time):
        """
        """
        raise NotImplementedError(
            "This service does not support getting historical price. "
            "Or rather it has no defined 'get_historical_price' method."
        )

    def get_transactions(self, crypto, address, confirmations=1):
        """
        Must be returned with the most recent transaction at the top.
        Returned is a list of dicts that have the following keys:

        `amount`: Number of units of currency moved. Always in base units (not satoshis).
        `date`: a datetime object of when the transaction was made.
        `txid`: The transaction ID, looks like a hash.
        `confirmations`: integer of the number of confirmations this transaction has on top of it.

        """
        raise NotImplementedError(
            "This service does not support getting historical transactions. "
            "Or rather it has no defined 'get_transactions' method."
        )

    def get_transactions_multi(self, crypto, addresses, confirmations=1):
        raise NotImplementedError(
            "This service does not support getting historical transactions by multiple addresses. "
            "Or rather it has no defined 'get_transactions_multi' method."
        )

    def get_single_transaction(self, crypto, txid):
        """
        Get detailed information about a single transaction.

        time - datetime, when this transaction was confirmed.
        block_hash - string, the id of the block this tx is confirmed in.
        block_number - integer, which block numbe this tx is confirmed in.
        hex - the entire tx encoded in hex format
        inputs - list of {address:, amount:}, amount is in satoshi
        outputs - list of {address:, amount:}, amount is in satoshi
        txid
        total_out
        confirmations

        """
        raise NotImplementedError(
            "This service does not support getting single transactions. "
            "Or rather it has no defined 'get_single_transaction' method."
        )

    def get_single_transaction_multi(self, crypto, txids):
        raise NotImplementedError(
            "This service does not support getting single transactions by multiple txids. "
            "Or rather it has no defined 'get_single_transaction_multi' method."
        )


    def get_unspent_outputs(self, crypto, address):
        """
        Default implmentation of this function that uses get_transaction
        Subclasses should overwrite this with a direct call to get utxo (if applicable)

        Returned will be a list of dictionaries, the keys will be:

        required:

        `output` - the big endian tx hash, followed by a colon, then the tx index. (for pybitcointools support)
        `address` - the address passed in (for pybitcointools support)
        `amount` - int, satoshi amount of the input
        `scriptPubKey` - string
        `txid` - string
        `vout` - integer

        optional:

        `confirmations` - how many confirmations this tx has so far.
        `scriptPubKey_asm` - string

        """
        raise NotImplementedError(
            "This service does not support getting unspent outputs. "
            "Or rather it has no defined 'get_unspent_outputs' method."
        )

    def get_unspent_outputs_multi(self, crypto, addresses):
        raise NotImplementedError(
            "This service does not support getting unspent outputs by multiple addresses. "
            "Or rather it has no defined 'get_unspent_outputs_multi' method."
        )

    def get_balance(self, crypto, address, confirmations=1):
        """
        Get the amount of coin in the address passed in.
        Always returns a single float.
        """
        raise NotImplementedError(
            "This service does not support getting address balances. "
            "Or rather it has no defined 'get_balance' method."
        )

    def get_balance_multi(self, crypto, addresses, confirmations=1):
        """
        Same as above, except addresses are passed in as a list instead of
        just a single string.
        """
        raise NotImplementedError(
            "This service does not support getting multiple address balances. "
            "Or rather it has no defined 'get_balance_multi' method."
        )

    def push_tx(self, crypto, tx_hex):
        """
        Push transaction to the miner network. Returns txid if done
        successfully.
        """
        raise NotImplementedError(
            "This service does not support pushing transactions to the network. "
            "Or rather it has no defined 'push_tx' method."
        )

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        """
        Get block based on either block height, block number or get the latest
        block. Only one of the previous arguments must be passed on.

        Returned is a dictionary object with the following keys:

        * required fields:

        block_number - int
        size - size of block
        time - datetime object of when the block was made
        hash - str (must be all lowercase)
        tx_count - int, the number of transactions included in thi block.

        * optional fields:

        confirmations - int
        sent_value - total value moved from all included transactions
        total_fees - total amount of tx fees from all included transactions
        mining_difficulty - what the difficulty was when this block was made.
        merkle_root - str (lower case)
        previous_hash - str (lower case)
        next_hash - str (lower case) (or `None` of its the latest block)
        """
        raise NotImplementedError(
            "This service does not support getting getting block data. "
            "Or rather it has no defined 'get_block' method."
        )

    def get_optimal_fee(self, crypto, tx_bytes):
        raise NotImplementedError(
            "This service does not support getting optimal fee. "
            "Or rather it has no defined 'get_optimal_fee' method."
        )


class AutoFallbackFetcher(object):
    """
    Calls a succession of services until one returns a value.
    """

    def __init__(self, services=None, verbose=False, responses=None, timeout=None, random_wait_seconds=0):
        """
        Each service class is instantiated here so the service instances stay
        in scope for the entire life of this object. This way the service
        objects can cache responses.
        """
        if not services:
            from moneywagon import ALL_SERVICES
            services = ALL_SERVICES

        self.services = []
        for ServiceClass in services:
            self.services.append(
                ServiceClass(verbose=verbose, responses=responses, timeout=timeout)
            )

        self.verbose = verbose
        self._successful_service = None # gets filled in after success
        self._failed_services = []
        self.random_wait_seconds = random_wait_seconds

    def _try_services(self, method_name, *args, **kwargs):
        """
        Try each service until one returns a response. This function only
        catches the bare minimum of exceptions from the service class. We want
        exceptions to be raised so the service classes can be debugged and
        fixed quickly.
        """
        if self.random_wait_seconds > 0:
            # for privacy... To avoid correlating addresses to same origin
            # only gets called before the first service call. Does not pause
            # before each and every call.
            pause_time = random.random() * self.random_wait_seconds
            if self.verbose:
                print("Pausing for: %.2f seconds" % pause_time)
            time.sleep(pause_time)

        for service in self.services:
            crypto = ((args and args[0]) or kwargs['crypto']).lower()
            address = kwargs.get('address', '').lower()
            fiat = kwargs.get('fiat', '').lower()

            if service.supported_cryptos and (crypto not in service.supported_cryptos):
                if self.verbose:
                    print("SKIP:", "%s not supported for %s" % (crypto, service.__class__.__name__))
                continue
            try:
                if self.verbose: print("* Trying:", service, crypto, "%s%s" % (address, fiat))
                ret =  getattr(service, method_name)(*args, **kwargs)
                self._successful_service = service
                return ret
            except (KeyError, IndexError, TypeError, ValueError,
                    requests.exceptions.Timeout, requests.exceptions.SSLError) as exc:
                # API has probably changed, therefore service class broken
                if self.verbose: print("FAIL:", service, exc.__class__.__name__, exc)
                self._failed_services.append({
                    'service': service,
                    'error': "%s %s" % (exc.__class__.__name__, exc)
                })
            except SkipThisService as exc:
                # service classes can raise this exception if for whatever reason
                # that service can't return a response, but maybe another one can.
                if self.verbose: print("SKIP:", exc.__class__.__name__, exc)
                self._failed_services.append({'service': service, 'error': "Skipped"})
            except NotImplementedError as exc:
                if self.verbose: print("SKIP:", exc.__class__.__name__, exc)
                self._failed_services.append({'service': service, 'error': "Not Implemented"})


        if not self._failed_services:
            raise NotImplementedError(
                "No Services defined for %s and %s" % (crypto, method_name)
            )

        if set(x['error'] for x in self._failed_services) == set(['Not Implemented']) and method_name.endswith("multi"):
            # some currencies may not have any multi functions defined, so retry
            # with private mode (which tries multiple services).
            raise RevertToPrivateMode("All services do not implement %s service" % method_name)

        failed_msg = ', '.join(
            ["{service.name} -> {error}".format(**x) for x in self._failed_services]
        )
        raise NoService(self.no_service_msg(*args, **kwargs) + "! Tried: " + failed_msg)

    def no_service_msg(self, *args, **kwargs):
        """
        This function is called when all Services have been tried and no value
        can be returned. It much take the same args and kwargs as in the method
        spefified in `method_name`. Returned is a string for the error message.
        It should say something informative.
        """
        return "All either skipped or failed."


def enforce_service_mode(services, FetcherClass, kwargs, modes):
    """
    Fetches the value according to the mode of execution desired.
    `FetcherClass` must be a class that is subclassed from AutoFallbackFetcher.
    `services` must be a list of Service classes.
    `kwargs` is a list of arguments used to make the service call, usually
      something like {crypto: 'btc', address: '1HwY...'} or
      {crypto: 'ltc', fiat: 'rur'}, (depends on the what FetcherClass.action takes)

    Modes can be:

         average = positive int. 1 by default. Takes average of n fetches.
         verbose = [True|False] False by default. Extra output.
         random = [True|False] False by default. Randomizes service order.
         paranoid = positive int. 1 by default. Redundant Fetching.
         fast = positive int. 0 by default. Return as soon as recieved first n results.

    """
    fast_level = modes.get('fast', 0)
    average_level = modes.get('average', 0)
    paranoid_level = modes.get('paranoid', 0)
    private_level = modes.get('private', 0)
    verbose = modes.get('verbose', False)
    timeout = modes.get('timeout', None)

    if modes.get('random', False):
        random.shuffle(services)

    if private_level  > 0:
        results = _do_private_mode(
            FetcherClass, services, kwargs, random_wait_seconds=private_level,
            verbose=verbose, timeout=timeout
        )
        return results

    elif average_level <= 1 and paranoid_level <= 1 and fast_level == 0:
        # only need to make 1 external call, no need for threading...
        fetcher = FetcherClass(services=services, verbose=verbose, timeout=timeout)
        consensus_results = fetcher.action(**kwargs)
        used_services = [fetcher._successful_service]

    elif average_level > 1:
        # instead of checking that all results are the same, we just return the average of all results.
        # mostly useful for non-blockchain operations like price and optimal fee.
        results = _get_results(
            FetcherClass, services, kwargs, num_results=average_level, verbose=verbose, timeout=timeout
        )

        to_compare, used_services = _prepare_consensus(FetcherClass, results)
        consensus_results = sum(to_compare) / len(to_compare)

    elif paranoid_level > 1:
        results = _get_results(
            FetcherClass, services, kwargs, num_results=paranoid_level, verbose=verbose, timeout=timeout
        )
        to_compare, used_services = _prepare_consensus(FetcherClass, results)

        if not len(set(to_compare)) == 1:
            ServiceDisagreement("Differing values returned: %s" % results)

        # if all values match, return any one (in this case the first one).
        # also return the list of all services that confirm this result.
        consensus_results = results[0][1]

    elif fast_level >= 1:
        raise Exception("Fast mode not yet working")
        results = _get_results(
            FetcherClass, services, kwargs, fast=fast_level, verbose=verbose
        )

    else:
        raise Exception("No mode specified.")

    if modes.get('report_services'):
        return used_services, consensus_results
    else:
        return consensus_results


def _prepare_consensus(FetcherClass, results):
    """
    Given a list of results, return a list that is simplified to make consensus
    determination possible. Returns two item tuple, first arg is simplified list,
    the second argument is a list of all services used in making these results.
    """
    # _get_results returns lists of 2 item list, first element is service, second is the returned value.
    # when determining consensus amoung services, only take into account values returned.
    if hasattr(FetcherClass, "strip_for_consensus"):
        to_compare = [
            FetcherClass.strip_for_consensus(value) for (service, value) in results
        ]
    else:
        to_compare = [value for service, value in results]

    return to_compare, [service._successful_service for service, values in results]

def _get_results(FetcherClass, services, kwargs, num_results=None, fast=0, verbose=False, timeout=None):
    """
    Does the fetching in multiple threads of needed. Used by paranoid and fast mode.
    """
    results = []

    if not num_results or fast:
        num_results = len(services)

    with futures.ThreadPoolExecutor(max_workers=len(services)) as executor:
        fetches = {}
        for service in services[:num_results]:
            tail = [x for x in services if x is not service]
            random.shuffle(tail)
            srv = FetcherClass(services=[service] + tail, verbose=verbose, timeout=timeout)
            fetches[executor.submit(srv.action, **kwargs)] = srv

        if fast == 1:
            raise NotImplementedError
            # ths code is a work in progress. futures.FIRST_COMPLETED works differently than I thought...
            to_iterate, still_going = futures.wait(fetches, return_when=futures.FIRST_COMPLETED)
            for x in still_going:
                try:
                    x.result(timeout=1.001)
                except futures._base.TimeoutError:
                    pass

        elif fast > 1:
            raise Exception("fast level greater than 1 not yet implemented")
        else:
            to_iterate = futures.as_completed(fetches)

        for future in to_iterate:
            service = fetches[future]
            results.append([service, future.result()])

    return results

def _do_private_mode(FetcherClass, services, kwargs, random_wait_seconds, timeout, verbose):
    """
    Private mode is only applicable to address_balance, unspent_outputs, and
    historical_transactions. There will always be a list for the `addresses`
    argument. Each address goes to a random service. Also a random delay is
    performed before the external fetch for improved privacy.
    """
    addresses = kwargs.pop('addresses')
    results = {}

    with futures.ThreadPoolExecutor(max_workers=len(addresses)) as executor:
        fetches = {}
        for address in addresses:
            k = kwargs
            k['address'] = address
            random.shuffle(services)
            srv = FetcherClass(
                services=services, verbose=verbose, timeout=timeout or 5.0,
                random_wait_seconds=random_wait_seconds
            )
            # address is returned because balance needs to be returned
            # attached to the address. Other methods (get_transaction, unspent_outputs, etc)
            # do not need to be indexed by address. (upstream they are stripped out)
            fetches[executor.submit(srv.action, **k)] = (srv, address)

        to_iterate = futures.as_completed(fetches)

        for future in to_iterate:
            service, address = fetches[future]
            results[address] = future.result()

    return results


def currency_to_protocol(amount):
    """
    Convert a string of 'currency units' to 'protocol units'. For instance
    converts 19.1 bitcoin to 1910000000 satoshis.

    Input is a float, output is an integer that is 1e8 times larger.

    It is hard to do this conversion because multiplying
    floats causes rounding nubers which will mess up the transactions creation
    process.

    examples:

    19.1 -> 1910000000
    0.001 -> 100000

    """
    if type(amount) in [float, int]:
        amount = "%.8f" % amount

    return int(amount.replace(".", '')) # avoiding float math

def get_optimal_services(crypto, type_of_service):
    from .crypto_data import crypto_data
    try:
        # get best services from curated list
        return crypto_data[crypto.lower()]['services'][type_of_service]
    except KeyError:
        raise ValueError("Invalid cryptocurrency symbol: %s" % crypto)

def get_magic_bytes(crypto):
    from .crypto_data import crypto_data
    try:
        pub_byte = crypto_data[crypto]['address_version_byte']
        priv_byte = crypto_data[crypto]['private_key_prefix']
        if priv_byte >= 128:
            priv_byte -= 128 #pybitcointools bug
        return pub_byte, priv_byte

    except KeyError:
        raise ValueError("Cryptocurrency symbol not found: %s" % crypto)



from moneywagon.services import Service, BitcoinAbe, BitpayInsight

class BlockStrap(Service):
    """
    Documentation here: http://docs.blockstrap.com/en/api/
    """
    #decomissioned
    service_id = 10
    domain = 'http://api.blockstrap.com'
    supported_cryptos = ['btc', 'ltc', 'drk', 'doge']
    api_homepage = "http://blockstrap.com/blog/deprecated/"

    def get_balance(self, crypto, address, confirmations=None):
        url = "%s/v0/%s/address/id/%s" % (self.domain, crypto, address)
        response = self.get_url(url).json()
        return response['data']['address']['balance'] / 1e8

    def push_tx(self, crypto, tx_hex):
        url = "%s/v0/%s/transaction/relay/%s" % (self.domain, crypto, tx_hex)
        return self.get_url(url).json()['data']['id']

    def get_transactions(self, crypto, address):
        url = "%s/v0/%s/address/transactions/%s" % (self.domain, crypto, address)
        txs = []
        for tx in self.get_url(url).json()['data']['address']['transactions']:
            s_amount = tx['tx_address_input_value'] or tx['tx_address_output_value'] * -1
            txs.append(dict(
                date=arrow.get(tx['block_time']).datetime,
                amount=s_amount / 1e8,
                confirmations=tx['confirmations'],
                txid=tx['id'],
            ))
        return txs

    def get_unspent_outputs(self, crypto, address):
        url = "%s/v0/%s/address/unspents/%s" % (self.domain, crypto, address)
        utxos = []
        for utxo in self.get_url(url).json()['data']['address']['transactions']:
            utxos.append(dict(
                amount=utxo['tx_address_value'],
                address=address,
                output="%s:%s" % (utxo['id'].lower(), utxo['tx_address_tx_pos']),
                confirmations=utxo['confirmations']
            ))
        return utxos

    def get_block(self, crypto, block_hash='', block_number='', latest=False):
        if block_hash:
            url = "%s/v0/%s/block/id/%s" % (self.domain, crypto, block_hash)
            b = "block"
        elif block_number:
            url = "%s/v0/%s/block/height/%s" % (self.domain, crypto, block_number)
            b = "blocks"
        elif latest:
            url = "%s/v0/%s/block/latest" % (self.domain, crypto)
            b = 'block'

        r = self.get_url(url).json()['data'][b]

        if block_number:
            r = r[0]

        return dict(
            block_number=r['height'],
            confirmations=r["confirmations"],
            time=arrow.get(r['time_display']).datetime,
            sent_value=r["output_value"] / 1e8,
            total_fees=r["fees"] / 1e8,
            size=r['size'],
            hash=r['id'].lower(),
            merkle_root=r['merkel_root'].lower(),
            previous_hash=r['prev_block_id'].lower(),
            next_hash=r['next_block_id'].lower(),
            txids=sorted(x['id'] for x in r['transactions']),
            tx_count=len(r['transactions'])
        )

class LitecoinAbe(BitcoinAbe):
    service_id = 16
    supported_cryptos = ['ltc']
    base_url = "http://bitcoin-abe.info/chain/Litecoin"


class NamecoinAbe(BitcoinAbe):
    service_id = 17
    supported_cryptos = ['nmc']
    base_url = "http://bitcoin-abe.info/chain/Namecoin"

class CoinSwap(Service):
    service_id = 26
    def get_current_price(self, crypto, fiat):
        chunk = ("%s/%s" % (crypto, fiat)).upper()
        url = "https://api.coin-swap.net/market/stats/%s" % chunk
        response = self.get_url(url).json()
        return float(response['lastprice']), 'coin-swap'


class ExCoIn(Service):
    service_id = 27
    # decommissioned
    def get_current_price(self, crypto, fiat):
        url = "https://api.exco.in/v1/exchange/%s/%s" % (fiat, crypto)
        response = self.get_url(url).json()
        return float(response['last_price']), 'exco.in'

class TheBitInfo(BitpayInsight):
    service_id = 29
    supported_cryptos = ['btc']
    domain = "insight.thebit.info"
    name = "TheBit.info"

class FTCc(BitpayInsight):
    service_id = 34
    supported_cryptos = ['ftc']
    domain = "block.ftc-c.com"
    name = "Feathercoin China"
